"""Unit tests for reviewboard.certs.cert.Certificate.

Version Added:
    6.0
"""

from __future__ import annotations

import os
import tempfile

import kgb

from reviewboard.certs.cert import CertDataFormat, CertificateBundle
from reviewboard.certs.errors import (CertificateNotFoundError,
                                      CertificateStorageError,
                                      InvalidCertificateFormatError)
from reviewboard.certs.tests.testcases import (CertificateTestCase,
                                               TEST_CERT_BUNDLE_PEM)


class CertificateBundleTests(kgb.SpyAgency, CertificateTestCase):
    """Unit tests for CertificateBundle.

    Version Added:
        6.0
    """
    def test_init_without_slug(self) -> None:
        """Testing CertificateBundle.__init__ with name not slug"""
        message = (
            'The certificate bundle name "test 123" must be in "slug" '
            'format (using characters "a-z", "0-9", "-").'
        )

        with self.assertRaisesMessage(ValueError, message):
            CertificateBundle(name='test 123',
                              bundle_data=b'...')

    def test_create_from_file(self) -> None:
        """Testing CertificateBundle.create_from_file"""
        fd, path = tempfile.mkstemp()
        os.write(fd, TEST_CERT_BUNDLE_PEM)
        os.close(fd)

        try:
            bundle = CertificateBundle.create_from_file(
                name='my-cert-bundle',
                path=path)
        finally:
            os.unlink(path)

        self.assertEqual(bundle.bundle_data, TEST_CERT_BUNDLE_PEM)
        self.assertEqual(bundle.name, 'my-cert-bundle')
        self.assertEqual(bundle.data_format, CertDataFormat.PEM)

    def test_create_from_file_with_not_found(self) -> None:
        """Testing CertificateBundle.create_from_file with not found"""
        message = 'The SSL/TLS CA bundle was not found.'

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            CertificateBundle.create_from_file(
                name='my-cert-bundle',
                path='/rb-tests-xxx/bad/')

    def test_create_from_file_with_ioerror(self) -> None:
        """Testing CertificateBundle.create_from_file with IOError"""
        self.spy_on(os.path.exists, op=kgb.SpyOpReturn(True))

        message = (
            r'Error loading SSL/TLS CA bundle file\. Administrators can '
            r'find details in the Review Board server logs \(error ID '
            r'[a-z0-9-]+\)\.'
        )

        with self.assertRaisesRegex(CertificateStorageError, message):
            CertificateBundle.create_from_file(
                name='my-cert-bundle',
                path='/rb-tests-xxx/bad/')

    def test_create_from_file_with_invalid(self) -> None:
        """Testing CertificateBundle.create_from_file with invalid format"""
        fd, path = tempfile.mkstemp()
        os.write(fd, b'XXX')
        os.close(fd)

        try:
            with self.assertRaises(InvalidCertificateFormatError):
                CertificateBundle.create_from_file(
                    name='my-cert-bundle',
                    path=path)
        finally:
            os.unlink(path)

    def test_write_bundle_file(self) -> None:
        """Testing CertificateBundle.write_bundle_file"""
        bundle = CertificateBundle(name='my-cert-bundle',
                                   bundle_data=TEST_CERT_BUNDLE_PEM)

        fd, path = tempfile.mkstemp()
        os.close(fd)

        try:
            bundle.write_bundle_file(path)

            with open(path, 'rb') as fp:
                self.assertEqual(fp.read(), TEST_CERT_BUNDLE_PEM)
        finally:
            os.unlink(path)

    def test_write_bundle_file_with_ioerror(self) -> None:
        """Testing CertificateBundle.write_bundle_file with IOError"""
        bundle = CertificateBundle(name='my-cert-bundle',
                                   bundle_data=TEST_CERT_BUNDLE_PEM)

        message = (
            r'Error writing SSL/TLS CA bundle file\. Administrators can '
            r'find details in the Review Board server logs \(error ID '
            r'[a-z0-9-]+\)\.'
        )

        with self.assertRaisesRegex(CertificateStorageError, message):
            bundle.write_bundle_file('/rb-tests-xxx/bad/')



"""Decorators for checking Local Site access."""

from django.http import HttpResponseRedirect, Http404
from django.shortcuts import render
from django.urls import reverse
from djblets.util.decorators import simple_decorator


@simple_decorator
def check_local_site_access(view_func):
    """Checks if a user has access to a Local Site.

    This checks whether or not the logged-in user is either a member of
    a Local Site or if the user otherwise has access to it.
    given local site. If not, this shows a permission denied page.
    """
    def _check(request, local_site_name=None, *args, **kwargs):
        if local_site_name:
            if not request.local_site:
                raise Http404

            local_site = request.local_site

            if not local_site.is_accessible_by(request.user):
                if local_site.public or request.user.is_authenticated:
                    return render(request=request,
                                  template_name='permission_denied.html',
                                  status=403)
                else:
                    return HttpResponseRedirect(
                        '%s?next=%s'
                        % (reverse('login'), request.get_full_path()))
        else:
            local_site = None

        return view_func(request, local_site=local_site, *args, **kwargs)

    return _check


@simple_decorator
def check_localsite_admin(view_func):
    """Checks if a user is an admin on a Local Site.

    This checks whether or not the logged-in user is marked as an admin for the
    given local site. If not, this shows a permission denied page.
    """
    def _check(request, local_site_name=None, *args, **kwargs):
        if local_site_name:
            if not request.local_site:
                raise Http404

            local_site = request.local_site

            if not local_site.is_mutable_by(request.user):
                return render(request=request,
                              template_name='permission_denied.html',
                              status=403)
        else:
            local_site = None

        return view_func(request,
                         local_site_name=local_site_name,
                         local_site=local_site,
                         *args, **kwargs)

    return _check



"""Root diff comments API resource.

Version Added:
    5.0
"""

from django.db.models import Q
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.fields import (BooleanFieldType,
                                   DateTimeFieldType,
                                   IntFieldType,
                                   StringFieldType)

from reviewboard.accounts.models import User
from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.reviews.models import Comment
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.resources.base_diff_comment import \
    BaseDiffCommentResource


class RootDiffCommentResource(BaseDiffCommentResource):
    """Provide information on diff comments.

    This is a top level endpoint that allows you to list and query all
    diff comments in the system, across different review requests.

    Version Changed:
        5.0.4:
        This was formally named ``diff_comments`` on the root resource, but
        had to be renamed to ``all_diff_comments`` in order to work around a
        conflict with URI templates.

    Version Added:
        5.0
    """

    added_in = '5.0'
    allowed_methods = ('GET',)
    model = Comment
    uri_object_key = None

    link_name = 'all_diff_comments'
    uri_template_name = 'all_diff_comment'

    mimetype_list_resource_name = 'review-diff-comments'

    @webapi_check_local_site
    def get_queryset(self, request, is_list=False, *args, **kwargs):
        """Return a queryset for Comment models.

        By default, this returns all comments that are accessible to
        the requester.

        The queryset can be further filtered by one or more arguments
        in the URL. These are listed in :py:meth:`get_list`.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            is_list (bool, unused):
                Whether the coming HTTP request is request for list resources.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for Comment models.
        """
        local_site = request.local_site
        q = Q()

        if 'file-diff-id' in request.GET:
            q &= Q(filediff=request.GET.get('file-diff-id'))

        if 'interdiff-revision' in request.GET:
            q &= Q(('interfilediff__diffset__revision',
                   request.GET.get('interdiff-revision')))

        if 'is-reply' in request.GET:
            val = request.GET.get('is-reply')
            q &= Q(reply_to_id__isnull=(not val))

        if 'last-updated-from' in request.GET:
            q &= Q(timestamp__gte=request.GET.get('last-updated-from'))

        if 'last-updated-to' in request.GET:
            q &= Q(timestamp__lt=request.GET.get('last-updated-to'))

        if 'review-id' in request.GET:
            q &= Q(review=request.GET.get('review-id'))

        if 'review-request-id' in request.GET:
            review_request_id = request.GET.get('review-request-id')

            if local_site is None:
                q &= Q(review__review_request=review_request_id)
            else:
                q &= (Q(review__review_request__local_id=review_request_id) &
                      Q(review__review_request__local_site=local_site))

        if 'user' in request.GET:
            user = list((
                User.objects
                .filter(username=request.GET.get('user'))
                .values_list('pk', flat=True)
            ))

            if user:
                q &= Q(review__user=user[0])
            else:
                return self.model.objects.none()

        return self.model.objects.accessible(request.user,
                                             extra_query=q,
                                             local_site=local_site)

    @webapi_request_fields(
        optional={
            'file-diff-id': {
                'type': StringFieldType,
                'description': 'The file diff ID that the diff comments must '
                               'be belonged to.',
            },
            'interdiff-revision': {
                'type': IntFieldType,
                'description': 'The second revision in an interdiff revision '
                               'range. The comments will be limited to this '
                               'range.',
            },
            'is-reply': {
                'type': BooleanFieldType,
                'description': 'Determine whether to return diff comments '
                               'that are replies or not.',
            },
            'last-updated-from': {
                'type': DateTimeFieldType,
                'description': "The earliest date/time the diff comments "
                               "could be last updated. This is compared "
                               "against the diff comment's "
                               "``timestamp`` field. This must be a valid "
                               ":term:`date/time format`.",
            },
            'last-updated-to': {
                'type': DateTimeFieldType,
                'description': "The date/time that all diff comments must "
                               "be last updated before. This is compared "
                               "against the diff comment's "
                               "``timestamp`` field. This must be a valid "
                               ":term:`date/time format`.",
            },
            'line': {
                'type': IntFieldType,
                'description': 'Filter for comments that start on a '
                               'particular line in the file.',
            },
            'review-id': {
                'type': StringFieldType,
                'description': 'The ID of the review that the diff comments '
                               'must belong to.',
            },
            'review-request-id': {
                'type': StringFieldType,
                'description': 'The ID of the review request that the diff '
                               'comments must belong to.',
            },
            'user': {
                'type': StringFieldType,
                'description': 'The username of the user that created the '
                               'diff comments.',
            },
        },
        allow_unknown=True
    )
    @augment_method_from(BaseDiffCommentResource)
    def get_list(self, *args, **kwargs):
        """Return the list of file attachment comments."""
        pass

    def serialize_object(self, obj, request=None, *args, **kwargs):
        """Serialize a diff comment.

        Args:
            obj (reviewboard.reviews.models.diff_comment.Comment):
                The diff comment to serialize.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The serialized diff comment.
        """
        result = super(RootDiffCommentResource, self).serialize_object(
            obj, request=request, *args, **kwargs)

        if not dvcs_feature.is_enabled(request=request):
            result.pop('base_filediff_id', None)

        return result


root_diff_comment_resource = RootDiffCommentResource()



"""Management command to clean up orphaned data.

Version Added:
    7.0
"""

from __future__ import annotations

import gc
import sys
from typing import Iterator, Optional, TYPE_CHECKING

import tqdm
from django.core.management.base import BaseCommand
from django.db import reset_queries
from django.utils.translation import (
    gettext as _,
    ngettext as N_)

from reviewboard.attachments.models import (FileAttachment,
                                            FileAttachmentHistory)
from reviewboard.changedescs.models import ChangeDescription
from reviewboard.diffviewer.models import DiffSet, DiffSetHistory
from reviewboard.reviews.models import Screenshot

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from django.db.models import Model, QuerySet
    from typing_extensions import TypeVar

    _ModelT = TypeVar('_ModelT', bound=Model)


class Command(BaseCommand):
    """Management command to clean up orphaned data.

    Version Added:
        7.0
    """

    help = _('Deletes orphaned file attachments')

    def add_arguments(
        self,
        parser: ArgumentParser,
    ) -> None:
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            '--show-counts-only',
            action='store_true',
            dest='show_counts',
            default=False,
            help=_('Show the number of orphaned items expected to be '
                   'deleted, but do not actually delete any data.'))
        parser.add_argument(
            '--no-progress',
            action='store_false',
            dest='show_progress',
            default=True,
            help=_("Don't show progress information or totals while "
                   "performing the clean up."))

    def handle(
        self,
        *,
        show_counts: bool = False,
        show_progress: bool = True,
        **options,
    ) -> None:
        """Handle the command.

        Args:
            show_counts (bool):
                Whether to only show the counts of objects.

            show_progress (bool):
                Whether to show progress information.

            **options (dict, unused):
                Any additional parsed command-line options.

        Raises:
            django.core.management.CommandError:
                An error occurred while cleaning up file attachments.
        """
        attachments = FileAttachment.objects.filter(
            user=None,
            drafts=None,
            inactive_drafts=None,
            review_request=None,
            inactive_review_request=None)

        attachment_histories = FileAttachmentHistory.objects.filter(
            review_request=None)

        screenshots = Screenshot.objects.filter(
            drafts=None,
            inactive_drafts=None,
            review_request=None,
            inactive_review_request=None)

        changedescs = ChangeDescription.objects.filter(
            review_request=None,
            review_request_draft=None)

        diffset_histories = DiffSetHistory.objects.filter(
            review_request=None)

        diffsets = DiffSet.objects.filter(
            history_id=None,
            review_request_draft=None)

        n_attachments = attachments.count()
        n_attachment_histories = attachment_histories.count()
        n_screenshots = screenshots.count()
        n_changedescs = changedescs.count()
        n_diffset_histories = diffset_histories.count()
        n_diffsets = diffsets.count()

        if (n_attachments == 0 and
            n_attachment_histories == 0 and
            n_screenshots == 0 and
            n_changedescs == 0 and
            n_diffset_histories == 0 and
            n_diffsets == 0):
            self.stdout.write(_('No orphaned data found.\n'))
            return

        if show_counts:
            self.stdout.write(
                N_('%d FileAttachment object\n',
                   '%d FileAttachment objects\n',
                   n_attachments)
                % n_attachments)
            self.stdout.write(
                N_('%d FileAttachmentHistory object\n',
                   '%d FileAttachmentHistory object\n',
                   n_attachment_histories)
                % n_attachment_histories)
            self.stdout.write(
                N_('%d Screenshot object\n',
                   '%d Screenshot objects\n',
                   n_screenshots)
                % n_screenshots)
            self.stdout.write(
                N_('%d ChangeDescription object\n',
                   '%d ChangeDescription objects\n',
                   n_changedescs)
                % n_changedescs)
            self.stdout.write(
                N_('%d DiffSetHistory object\n',
                   '%d DiffSetHistory objects\n',
                   n_diffset_histories)
                % n_diffset_histories)
            self.stdout.write(
                N_('%d DiffSet object\n',
                   '%d DiffSet objects\n',
                   n_diffsets)
                % n_diffsets)
            return

        if show_progress:
            self.stdout.write(_('Deleting orphaned objects:\n'))

        # We explicitly delete file attachments before file attachment
        # histories. Deleting the histories would cause the file attachments to
        # get deleted in a cascade, which might theoretically put more load on
        # the DB than we want. This does mean we end up with some extra queries
        # to update the histories as file attachments get deleted, but that's
        # better than potentially timing out on a giant cascade.

        self._delete_in_batches(
            queryset=attachments,
            total_objects=n_attachments,
            show_progress=show_progress,
            description=_('FileAttachment'))

        self._delete_in_batches(
            queryset=attachment_histories,
            total_objects=n_attachment_histories,
            show_progress=show_progress,
            description=_('FileAttachmentHistory'))

        self._delete_in_batches(
            queryset=screenshots,
            total_objects=n_screenshots,
            show_progress=show_progress,
            description=_('Screenshot'))

        self._delete_in_batches(
            queryset=changedescs,
            total_objects=n_changedescs,
            show_progress=show_progress,
            description=_('ChangeDescription'))

        self._delete_in_batches(
            queryset=diffset_histories,
            total_objects=n_diffset_histories,
            show_progress=show_progress,
            description=_('DiffSetHistory'))

        self._delete_in_batches(
            queryset=diffsets,
            total_objects=n_diffsets,
            show_progress=show_progress,
            description=_('DiffSet'))

        if show_progress:
            self.stdout.write(_('Done.\n'))

    def _delete_in_batches(
        self,
        *,
        queryset: QuerySet[_ModelT],
        total_objects: int,
        show_progress: bool,
        description: str,
    ) -> None:
        """Delete the given objects in batches.

        Args:
            queryset (django.db.models.QuerySet):
                The queryset of objects to delete.

            total_objects (int):
                The total number of objects in the queryset.

            show_progress (bool):
                Whether to show progress.

            description (str):
                The description to show, when showing progress.
        """
        if total_objects == 0:
            # Skip the output entirely.
            return

        # If we're operating with a real stdout, we want to let tqdm connect
        # directly to it, rather than using django's OutputWrapper.
        if self.stdout._out is sys.stdout:
            file = None
        else:
            file = self.stdout

        t = tqdm.tqdm(
            desc=description,
            disable=not show_progress,
            file=file,
            total=total_objects,
            bar_format='{desc} {bar} [{n_fmt}/{total_fmt}]',
            ncols=80)

        for batch in self._iter_batches(queryset, total_objects=total_objects):
            n = batch.count()

            batch.delete()

            t.update(n)

        t.close()

    def _iter_batches(
        self,
        queryset: QuerySet[_ModelT],
        *,
        total_objects: Optional[int] = None,
        batch_size: int = 50,
    ) -> Iterator[QuerySet[_ModelT]]:
        """Iterate through items in a queryset, yielding batches.

        This will gather up to a specified number of items from a queryset
        at a time, process them into batches of a specified size, and yield
        them.

        After each set of objects is fetched from the database, garbage
        collection will be forced and stored queries reset, in order to
        reduce memory usage.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset to execute for fetching objects.

            total_objects (int, optional):
                The total number of objects. If this is ``None``, it will be
                determined from ``queryset.count()``.

            batch_size (int, optional):
                The maximum number of objects to yield per batch.

        Yields:
            django.db.models.query.QuerySet:
            A QuerySet containing a batch of items to process. This will never
            be larger than ``batch_size``.
        """
        if total_objects is None:
            total_objects = queryset.count()

        object_pks = list(queryset.values_list('pk', flat=True))

        for i in range(0, total_objects, batch_size):
            batch_pks = object_pks[i:i + batch_size]
            yield queryset.model.objects.filter(pk__in=batch_pks)

            # Do all we can to limit the memory usage by resetting any
            # stored queries (if DEBUG is True), and force garbage
            # collection of anything we may have from processing an object.
            reset_queries()
            gc.collect()



from django import forms
from django.contrib.auth.models import User
from django.test import RequestFactory
from djblets.forms.fields import ConditionsField

from reviewboard.integrations.base import Integration, get_integration_manager
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.conditions import ReviewRequestConditionChoices
from reviewboard.reviews.models import Group
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class MyIntegration(Integration):
    integration_id = 'my-integration'


class MyConfigForm(IntegrationConfigForm):
    my_conditions = ConditionsField(ReviewRequestConditionChoices)
    group = forms.ModelChoiceField(queryset=Group.objects.order_by('pk'))

    def serialize_group_field(self, group):
        return group.name


class IntegrationConfigFormTests(TestCase):
    """Unit tests for reviewboard.integrations.forms.IntegrationConfigForm."""

    def setUp(self):
        super(IntegrationConfigFormTests, self).setUp()

        self.local_site_1 = LocalSite.objects.create(name='site1')
        self.local_site_2 = LocalSite.objects.create(name='site2')

        self.manager = get_integration_manager()
        self.integration = \
            self.manager.register_integration_class(MyIntegration)

        self.request = RequestFactory().request()
        self.request.user = User.objects.create(username='test-user')

        self.local_site_1_group = self.create_review_group(
            name='local-site-1-group',
            local_site=self.local_site_1)

        self.local_site_2_group = self.create_review_group(
            name='local-site-2-group',
            local_site=self.local_site_2)

        self.global_site_group = self.create_review_group(
            name='global-site-group')

    def tearDown(self):
        super(IntegrationConfigFormTests, self).tearDown()

        self.manager.unregister_integration_class(MyIntegration)

    def test_without_localsite(self):
        """Testing IntegrationConfigForm without a LocalSite"""
        # Make sure the initial state and querysets are what we expect on init.
        form = MyConfigForm(integration=self.integration,
                            request=self.request)

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group,
                          self.local_site_2_group,
                          self.global_site_group])
        self.assertNotIn('local_site',
                         form.fields['my_conditions'].choice_kwargs)

        # Now test what happens when it's been fed data and validated.
        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.global_site_group.pk],
                'group': self.global_site_group.pk,
            })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group,
                          self.local_site_2_group,
                          self.global_site_group])
        self.assertNotIn('local_site',
                         form.fields['my_conditions'].choice_kwargs)

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group,
                          self.local_site_2_group,
                          self.global_site_group])

        config = form.save()
        self.assertIsNone(config.local_site)
        self.assertEqual(config.settings['group'], 'global-site-group')

        condition_set = config.settings['my_conditions']
        self.assertEqual(list(condition_set.conditions[0].value),
                         [self.global_site_group])

    def _test_init_with_limit_to_local_site(self):
        """Testing IntegrationConfigForm initialization with
        limit_to_local_site
        """
        LocalSite.objects.create(name='local-site-1')
        local_site = LocalSite.objects.create(name='local-site-2')

        form = MyConfigForm(integration=self.integration,
                            request=self.request,
                            limit_to_local_site=local_site)

        self.assertNotIn('local_site', form.fields)
        self.assertEqual(form.limited_to_local_site, local_site)
        self.assertEqual(
            form.fields['my_conditions'].choice_kwargs,
            {
                'local_site': local_site,
            })

    def test_without_localsite_and_instance(self):
        """Testing IntegrationConfigForm without a LocalSite and editing
        instance
        """
        config = IntegrationConfig.objects.create(
            integration_id=self.integration.integration_id)

        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            instance=config,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.global_site_group.pk],
                'group': self.global_site_group.pk,
            })

        self.assertTrue(form.is_valid())

        new_config = form.save()
        self.assertEqual(config.pk, new_config.pk)
        self.assertIsNone(new_config.local_site)

    def test_without_localsite_and_with_local_site_values(self):
        """Testing IntegrationConfigForm without a LocalSite and with field
        values on a LocalSite
        """
        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.local_site_1_group.pk],
                'group': self.local_site_2_group.pk,
            })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'group': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
                'my_conditions': [
                    'There was an error with one of your conditions.'
                ],
            })
        self.assertEqual(
            form.fields['my_conditions'].widget.condition_errors,
            {
                0: 'Select a valid choice. 1 is not one of the available '
                   'choices.',
            })

    def test_with_limited_localsite(self):
        """Testing IntegrationConfigForm limited to a LocalSite"""
        form = MyConfigForm(integration=self.integration,
                            request=self.request,
                            limit_to_local_site=self.local_site_1)

        self.assertEqual(form.limited_to_local_site, self.local_site_1)
        self.assertNotIn('local_site', form.fields)
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group])
        self.assertEqual(
            form.fields['my_conditions'].choice_kwargs.get('local_site'),
            self.local_site_1)

    def test_with_limited_localsite_and_changing_site(self):
        """Testing IntegrationConfigForm limited to a LocalSite and changing
        LocalSite
        """
        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.local_site_1_group.pk],
                'group': self.local_site_1_group.pk,
                'local_site': self.local_site_2.pk,
            },
            limit_to_local_site=self.local_site_1)

        self.assertEqual(form.limited_to_local_site, self.local_site_1)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['local_site'], self.local_site_1)
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group])
        self.assertEqual(
            form.fields['my_conditions'].choice_kwargs.get('local_site'),
            self.local_site_1)

        new_config = form.save()
        self.assertEqual(new_config.local_site, self.local_site_1)

    def test_with_limited_localsite_and_compatible_instance(self):
        """Testing IntegrationConfigForm limited to a LocalSite and editing
        compatible instance
        """
        config = IntegrationConfig.objects.create(
            integration_id=self.integration.integration_id,
            local_site=self.local_site_1)

        # This should just simply not raise an exception.
        MyConfigForm(integration=self.integration,
                     instance=config,
                     request=self.request,
                     limit_to_local_site=self.local_site_1)

    def test_with_limited_localsite_and_incompatible_instance(self):
        """Testing IntegrationConfigForm limited to a LocalSite and editing
        incompatible instance
        """
        config1 = IntegrationConfig.objects.create(
            integration_id=self.integration.integration_id)
        config2 = IntegrationConfig.objects.create(
            integration_id=self.integration.integration_id,
            local_site=self.local_site_2)

        error_message = (
            'The provided instance is not associated with a LocalSite '
            'compatible with this form. Please contact support.'
        )

        with self.assertRaisesMessage(ValueError, error_message):
            MyConfigForm(integration=self.integration,
                         request=self.request,
                         instance=config1,
                         limit_to_local_site=self.local_site_1)

        with self.assertRaisesMessage(ValueError, error_message):
            MyConfigForm(integration=self.integration,
                         request=self.request,
                         instance=config2,
                         limit_to_local_site=self.local_site_1)

    def test_with_limited_localsite_and_invalid_values(self):
        """Testing IntegrationConfigForm limited to a LocalSite with values
        not on the LocalSite
        """
        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.local_site_2_group.pk],
                'group': self.global_site_group.pk,
            },
            limit_to_local_site=self.local_site_1)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'group': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
                'my_conditions': [
                    'There was an error with one of your conditions.'
                ],
            })
        self.assertEqual(
            form.fields['my_conditions'].widget.condition_errors,
            {
                0: 'Select a valid choice. 2 is not one of the available '
                   'choices.',
            })

    def test_with_localsite_in_data(self):
        """Testing IntegrationConfigForm with a LocalSite in form data"""
        # Make sure the initial state and querysets are what we expect on init.
        form = MyConfigForm(integration=self.integration,
                            request=self.request)

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group,
                          self.local_site_2_group,
                          self.global_site_group])
        self.assertNotIn('local_site',
                         form.fields['my_conditions'].choice_kwargs)

        # Now test what happens when it's been fed data and validated.
        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.local_site_1_group.pk],
                'group': self.local_site_1_group.pk,
                'local_site': self.local_site_1.pk,
            })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group,
                          self.local_site_2_group,
                          self.global_site_group])
        self.assertNotIn('local_site',
                         form.fields['my_conditions'].choice_kwargs)

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['group'].queryset),
                         [self.local_site_1_group,
                          self.local_site_2_group,
                          self.global_site_group])
        self.assertNotIn('local_site',
                         form.fields['my_conditions'].choice_kwargs)

        new_config = form.save()
        self.assertEqual(new_config.local_site, self.local_site_1)
        self.assertEqual(new_config.settings['group'], 'local-site-1-group')

        condition_set = new_config.settings['my_conditions']
        self.assertEqual(list(condition_set.conditions[0].value),
                         [self.local_site_1_group])

    def test_with_localsite_in_data_and_instance(self):
        """Testing IntegrationConfigForm with a LocalSite in form data and
        editing instance
        """
        config = IntegrationConfig.objects.create(
            integration_id=self.integration.integration_id)

        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.local_site_1_group.pk],
                'group': self.local_site_1_group.pk,
                'local_site': self.local_site_1.pk,
            },
            instance=config)
        self.assertTrue(form.is_valid())

        new_config = form.save()
        self.assertEqual(config.pk, new_config.pk)
        self.assertEqual(new_config.local_site, self.local_site_1)

    def test_with_localsite_in_data_and_invalid_values(self):
        """Testing IntegrationConfigForm with a LocalSite in form data and
        values not on the LocalSite
        """
        form = MyConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'name': 'Test',
                'my_conditions_last_id': '0',
                'my_conditions_mode': 'all',
                'my_conditions_choice[0]': 'review-groups',
                'my_conditions_operator[0]': 'contains-any',
                'my_conditions_value[0]': [self.local_site_2_group.pk],
                'group': self.global_site_group.pk,
                'local_site': self.local_site_1.pk,
            })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'group': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
                'my_conditions': [
                    'There was an error with one of your conditions.'
                ],
            })
        self.assertEqual(
            form.fields['my_conditions'].widget.condition_errors,
            {
                0: 'Select a valid choice. 2 is not one of the available '
                   'choices.',
            })



from django.urls import include, path
from djblets.urls.resolvers import DynamicURLResolver


dynamic_urls = DynamicURLResolver()


urlpatterns = [
    path('repos/<int:repository_id>/', include([dynamic_urls])),
]



"""Unit tests for reviewboard.admin.form_widgets.RelatedRepositoryWidget."""

from django import forms

from reviewboard.admin.form_widgets import RelatedRepositoryWidget
from reviewboard.scmtools.models import Repository
from reviewboard.testing.testcase import TestCase


class TestForm(forms.Form):
    """A Test Form with a field that contains a RelatedRepositoryWidget."""
    my_multiselect_field = forms.ModelMultipleChoiceField(
        queryset=Repository.objects.filter(visible=True).order_by('name'),
        label=('Repositories'),
        required=False,
        widget=RelatedRepositoryWidget())


class LocalSiteTestForm(forms.Form):
    """A Test Form with a field that contains a RelatedRepositoryWidget.

    The RelatedRepositoryWidget is defined to have a local_site_name.
    """
    my_multiselect_field = forms.ModelMultipleChoiceField(
        queryset=Repository.objects.filter(visible=True).order_by('name'),
        label=('Repositories'),
        required=False,
        widget=RelatedRepositoryWidget(local_site_name='supertest'))


class SingleValueTestForm(forms.Form):
    """A Test Form with a field that contains a RelatedRepositoryWidget.

    RelatedRepositoryWidget is defined as setting multivalued to False.
    """
    my_select_field = forms.ModelMultipleChoiceField(
        queryset=Repository.objects.filter(visible=True).order_by('name'),
        label=('Repositories'),
        required=False,
        widget=RelatedRepositoryWidget(multivalued=False))


class RelatedRepositoryWidgetTests(TestCase):
    """Unit tests for RelatedRepositoryWidget."""

    fixtures = ['test_scmtools']

    def test_render_empty(self):
        """Testing RelatedRepositoryWidget.render with no initial data"""
        my_form = TestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Repositories',
            [],
            {'id': 'repositories'})

        self.assertHTMLEqual(
            """<input id="repositories" name="Repositories" type="hidden" />

            <script>
            $(function() {
                var view = new RB.RelatedRepoSelectorView({
                    $input: $('#repositories'),
                    "initialOptions": [],

                    "multivalued": true,
                    "showInvisible": true
                }).render();
            });
            </script>""",
            html)

    def test_render_with_data(self):
        """Testing RelatedRepositoryWidget.render with initial data"""
        test_repo_1 = self.create_repository(name='repo1')
        test_repo_2 = self.create_repository(name='repo2')
        test_repo_3 = self.create_repository(name='repo3')

        my_form = TestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Repositories',
            [test_repo_1.pk, test_repo_2.pk, test_repo_3.pk],
            {'id': 'repositories'})

        self.assertHTMLEqual(
            """<input id="repositories" name="Repositories" type="hidden"
                      value="1,2,3" />

            <script>
            $(function() {
                var view = new RB.RelatedRepoSelectorView({
                    $input: $('#repositories'),
                    "initialOptions": [{"id": 1, "name": "repo1"},
                    {"id": 2, "name": "repo2"},
                    {"id": 3, "name": "repo3"}],

                    "multivalued": true,
                    "showInvisible": true
                }).render();
            });
            </script>""",
            html)

    def test_render_with_local_site(self):
        """Testing RelatedRepositoryWidget.render with a local site defined"""
        my_form = LocalSiteTestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Repositories',
            [],
            {'id': 'repositories'})

        self.assertIn('"localSitePrefix": "s/supertest/",', html)

    def test_value_from_datadict(self):
        """Testing RelatedRepositoryWidget.value_from_datadict"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {'repository': ['1', '2']},
                {},
                'repository'))

        self.assertEqual(['1', '2'], value)

    def test_value_from_datadict_single_value(self):
        """Testing RelatedRepositoryWidget.value_from_datadict with a single
        value
        """
        my_form = SingleValueTestForm()
        value = (
            my_form.fields['my_select_field']
            .widget
            .value_from_datadict(
                {'repository': ['1']},
                {},
                'repository'))

        self.assertEqual(value, ['1'])

    def test_value_from_datadict_with_no_data(self):
        """Testing RelatedRepositoryWidget.value_from_datadict with no data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {'repository': []},
                {},
                'repository'))

        self.assertEqual(value, [])

    def test_value_from_datadict_with_missing_data(self):
        """Testing RelatedRepositoryWidget.value_from_datadict with missing
        data
        """
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {},
                {},
                'repository'))

        self.assertIsNone(value)



"""Models for hosting service accounts."""

from __future__ import annotations

import logging
from typing import ClassVar, TYPE_CHECKING
from urllib.parse import urlparse

from django.db import models
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import JSONField

from reviewboard.certs.cert import Certificate
from reviewboard.certs.manager import cert_manager
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base import hosting_service_registry
from reviewboard.hostingsvcs.errors import MissingHostingServiceError
from reviewboard.hostingsvcs.managers import HostingServiceAccountManager
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django.contrib.auth.models import User

    from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
    from reviewboard.scmtools.certs import Certificate as LegacyCertificate


logger = logging.getLogger(__name__)


class HostingServiceAccount(models.Model):
    """An account for a hosting service."""

    service_name = models.CharField(
        max_length=128,
        help_text=_('The ID of the hosting service that is associated '
                    'with this account.'))
    hosting_url = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        help_text=_('The hosting URL.'))
    username = models.CharField(
        max_length=128,
        help_text=_('The username of the account on the hosting service.'))
    data = JSONField(
        help_text=_('Account data specific to the hosting service. This '
                    'should generally not be changed.'))
    visible = models.BooleanField(
        default=True,
        help_text=_('Whether this account shows up as an option when '
                    'configuring a repository.'))
    local_site = models.ForeignKey(
        LocalSite,
        on_delete=models.CASCADE,
        related_name='hosting_service_accounts',
        verbose_name=_('Local site'),
        blank=True,
        null=True,
        help_text=_('The LocalSite to associate with this account.'))

    objects: ClassVar[HostingServiceAccountManager] = \
        HostingServiceAccountManager()

    def __str__(self) -> str:
        """Return a string representation of the hosting service account.

        Returns:
            str:
            A string representation of the object.
        """
        if self.hosting_url:
            # Show the hosting URL, so that users can distinguish between
            # the accounts across different self-hosted servers of a given
            # type.
            return f'{self.username} ({self.hosting_url})'
        else:
            return self.username

    @property
    def service(self) -> BaseHostingService:
        """The hosting service associated with this account.

        Type:
            reviewboard.hostingsvcs.base.hosting_service.BaseHostingService

        Raises:
            reviewboard.hostingsvcs.errors.MissingHostingServiceError:
                The hosting service could not be loaded from the registry.
        """
        if not hasattr(self, '_service'):
            service_name = self.service_name

            cls = hosting_service_registry.get_hosting_service(
                self.service_name)

            if cls:
                self._service = cls(self)
            else:
                logger.error('Failed to load hosting service %s for '
                             'repository %s.',
                             service_name, self.pk)

                raise MissingHostingServiceError(service_name)

        return self._service

    @property
    def is_authorized(self) -> bool:
        """Whether the service is authorized.

        Type:
            bool
        """
        service = self.service

        if service:
            return service.is_authorized()
        else:
            return False

    def is_accessible_by(
        self,
        user: User,
    ) -> bool:
        """Return whether or not the user has access to the account.

        The account is accessible by the user if the user has access to the
        local site.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            True if the user has access to the hosting service account.
        """
        return not self.local_site or self.local_site.is_accessible_by(user)

    def is_mutable_by(
        self,
        user: User,
    ) -> bool:
        """Return whether or not the user can modify or delete the account.

        The account is mutable by the user if the user is an administrator
        with proper permissions or the account is part of a LocalSite and
        the user has permissions to modify it.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            True if the user can modify the hosting service account.
        """
        return user.has_perm('hostingsvcs.change_hostingserviceaccount',
                             self.local_site)

    def accept_certificate(
        self,
        certificate: LegacyCertificate,
    ) -> None:
        """Accept the SSL certificate for the linked hosting URL.

        Deprecated:
            8.0:
            This has been replaced with
            :py:meth:`CertificateManager.add_certificate()
            <reviewboard.certs.manager.CertificateManager.add_certificate>`
            and will be removed in Review Board 10.

        Args:
            certificate (reviewboard.scmtools.certs.Certificate):
                The certificate to accept.

        Raises:
            ValueError:
                The certificate data did not include required fields.
        """
        RemovedInReviewBoard10_0Warning.warn(
            'HostingServiceAccount.accept_certificate() is deprecated and '
            'will be removed in Review Board 10. Use '
            'cert_manager.add_certificate() instead.'
        )

        cert_data = certificate.pem_data

        if not cert_data:
            raise ValueError('The certificate does not include a PEM-encoded '
                             'representation.')

        # Also register with the certificate manager so that it can be
        # checked against the main fingerprint storage.
        hosting_url = self.hosting_url
        hostname = certificate.hostname
        fingerprint = certificate.fingerprint
        port = 443

        if hosting_url:
            try:
                parsed = urlparse(hosting_url)

                if parsed.hostname:
                    hostname = parsed.hostname

                if parsed.port:
                    port = parsed.port
                elif parsed.scheme == 'http':
                    logger.error(
                        'Attempted to accept TLS/SSL certificate for HTTP '
                        'URL %s. This may be a programming error or a '
                        'misconfiguration with a server. A certificate '
                        'will not be added.',
                        hosting_url,
                    )

                    return
            except Exception as e:
                logger.exception(
                    'Unexpected error parsing the URL %s when accepting a '
                    'TLS/SSL certificate: %s',
                    hosting_url, e,
                )

                return

        if not hostname:
            logger.error(
                'Could not determine a hostname to use for the TLS/SSL '
                'certificate accepted for %s. A certificate will not be '
                'added.',
                hosting_url or '<unknown>',
            )

            return

        try:
            cert_manager.add_certificate(
                Certificate(
                    hostname=hostname,
                    port=port,
                    cert_data=cert_data.encode('ascii'),
                ),
                local_site=self.local_site,
            )
        except Exception as e:
            logger.error(
                'Failed to add SSL certificate for %s:%s '
                '(fingerprint %r): %s',
                hostname, port, fingerprint, e,
            )

    class Meta:
        """Metadata for the HostingServiceAccount model."""

        db_table = 'hostingsvcs_hostingserviceaccount'
        verbose_name = _('Hosting Service Account')
        verbose_name_plural = _('Hosting Service Accounts')



"""Unit tests for ActiveDirectoryBackend."""

import dns
import kgb
import ldap
from django.conf import settings
from django.contrib.auth.models import User
from django.test import override_settings

from reviewboard.accounts.backends import ActiveDirectoryBackend
from reviewboard.accounts.tests.test_ldap_auth_backend import TestLDAPObject
from reviewboard.testing import TestCase


class ActiveDirectoryBackendTests(kgb.SpyAgency, TestCase):
    """Unit tests for ActiveDirectoryBackend."""

    def setUp(self):
        super(ActiveDirectoryBackendTests, self).setUp()

        # These settings will get overridden on future test runs, since
        # they'll be reloaded from siteconfig.
        settings.AD_DOMAIN_CONTROLLER = 'ad1.example.com ad2.example.com:123'
        settings.AD_DOMAIN_NAME = 'example.com'
        settings.AD_RECURSION_DEPTH = -1

        self.backend = ActiveDirectoryBackend()

        self.spy_on(ldap.initialize,
                    call_fake=lambda uri, *args, **kwargs: TestLDAPObject(uri))

    def test_authenticate_with_valid_credentials(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                        }),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user',
                                         password='test-pass')

        self.assertIsNotNone(user)

        self.assertIsNotNone(user.pk)
        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertEqual(user.email, 'test-user@corp.example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@example.com',
            'test-pass')
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)

    def test_authenticate_with_valid_credentials_with_at(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        and "@" in username
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                        }),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user@test',
                                         password='test-pass')

        self.assertIsNotNone(user)

        self.assertIsNotNone(user.pk)
        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertEqual(user.email, 'test-user@corp.example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@test.example.com',
            'test-pass')
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=test,dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)

    def test_authenticate_with_valid_credentials_with_backslash(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        and "\\" in username
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                        }),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='test\\test-user',
                                         password='test-pass')

        self.assertIsNotNone(user)

        self.assertIsNotNone(user.pk)
        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertEqual(user.email, 'test-user@corp.example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@test.example.com',
            'test-pass')
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=test,dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)

    def test_authenticate_with_empty_attrs(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        and empty user attributes
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {}),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user',
                                         password='test-pass')

        self.assertIsNotNone(user)

        self.assertIsNotNone(user.pk)
        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'test-user')
        self.assertEqual(user.last_name, '')
        self.assertEqual(user.email, 'test-user@example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@example.com',
            'test-pass')
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)

    @override_settings(AD_GROUP_NAME='required')
    def test_authenticate_with_in_required_group(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        and in required group
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['dc=test,dc=corp,dc=example,dc=com'],
                    'call_fake': lambda *args, **kwargs: [
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                            'memberOf': [
                                b'cn=required,ou=Groups,dc=example,dc=com',
                            ],
                        }),
                    ],
                },
                {
                    'args': ['dc=example,dc=com'],
                    'kwargs': {
                        'filterstr': '(&(objectClass=group)(cn=required))',
                        'scope': ldap.SCOPE_SUBTREE,
                    },
                    'call_fake': lambda *args, **kwargs: [],
                },
            ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user@test.corp',
                                         password='test-pass')

        self.assertIsNotNone(user)

        self.assertIsNotNone(user.pk)
        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertEqual(user.email, 'test-user@corp.example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@test.corp.example.com',
            'test-pass')
        self.assertSpyCallCount(TestLDAPObject.search_s, 2)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=test,dc=corp,dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=group)(cn=required))',
            scope=ldap.SCOPE_SUBTREE)

    @override_settings(AD_GROUP_NAME='required')
    def test_authenticate_with_in_required_group_recurse(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        and in required group with recursion
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['dc=test,dc=corp,dc=example,dc=com'],
                    'call_fake': lambda *args, **kwargs: [
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                            'memberOf': [
                                b'cn=other,ou=Groups,dc=example,dc=com',
                            ],
                        }),
                    ],
                },
                {
                    'args': ['dc=example,dc=com'],
                    'kwargs': {
                        'filterstr': '(&(objectClass=group)(cn=other))',
                        'scope': ldap.SCOPE_SUBTREE,
                    },
                    'call_fake': lambda *args, **kwargs: [
                        ('cn=other,ou=Groups,dc=example,dc=com', {
                            'memberOf': [
                                b'cn=other2,ou=Groups,dc=example,dc=com',
                            ],
                        }),
                    ],
                },
                {
                    'args': ['dc=example,dc=com'],
                    'kwargs': {
                        'filterstr': '(&(objectClass=group)(cn=other2))',
                        'scope': ldap.SCOPE_SUBTREE,
                    },
                    'call_fake': lambda *args, **kwargs: [
                        ('cn=other2,ou=Groups,dc=example,dc=com', {
                            'memberOf': [
                                b'cn=other,ou=Groups,dc=example,dc=com',
                                b'cn=required,ou=Groups,dc=example,dc=com',
                            ],
                        }),
                    ],
                },
                {
                    'args': ['dc=example,dc=com'],
                    'kwargs': {
                        'filterstr': '(&(objectClass=group)(cn=required))',
                        'scope': ldap.SCOPE_SUBTREE,
                    },
                    'call_fake': lambda *args, **kwargs: [],
                },
            ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user@test.corp',
                                         password='test-pass')

        self.assertIsNotNone(user)

        self.assertIsNotNone(user.pk)
        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertEqual(user.email, 'test-user@corp.example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@test.corp.example.com',
            'test-pass')
        self.assertSpyCallCount(TestLDAPObject.search_s, 4)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=test,dc=corp,dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=group)(cn=other))',
            scope=ldap.SCOPE_SUBTREE)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=group)(cn=other2))',
            scope=ldap.SCOPE_SUBTREE)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=group)(cn=required))',
            scope=ldap.SCOPE_SUBTREE)

    @override_settings(AD_GROUP_NAME='required',
                       AD_RECURSION_DEPTH=1)
    def test_authenticate_with_required_group_recurse_limit_hit(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        and in required group recursion limit hit
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['dc=test,dc=corp,dc=example,dc=com'],
                    'call_fake': lambda *args, **kwargs: [
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                            'memberOf': [
                                b'cn=other,ou=Groups,dc=example,dc=com',
                            ],
                        }),
                    ],
                },
                {
                    'args': ['dc=example,dc=com'],
                    'kwargs': {
                        'filterstr': '(&(objectClass=group)(cn=other))',
                        'scope': ldap.SCOPE_SUBTREE,
                    },
                    'call_fake': lambda *args, **kwargs: [
                        ('cn=other,ou=Groups,dc=example,dc=com', {
                            'memberOf': [
                                b'cn=other2,ou=Groups,dc=example,dc=com',
                            ],
                        }),
                    ],
                },
            ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user@test.corp',
                                         password='test-pass')

        self.assertIsNone(user)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@test.corp.example.com',
            'test-pass')
        self.assertSpyCallCount(TestLDAPObject.search_s, 2)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=test,dc=corp,dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=group)(cn=other))',
            scope=ldap.SCOPE_SUBTREE)

    @override_settings(AD_GROUP_NAME='required')
    def test_authenticate_with_not_in_required_group(self):
        """Testing ActiveDirectoryBackend.authenticate with valid credentials
        and not in required group
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject)
        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['dc=test,dc=corp,dc=example,dc=com'],
                    'kwargs': {
                        'filterstr': ('(&(objectClass=user)'
                                      '(sAMAccountName=test-user))'),
                        'scope': ldap.SCOPE_SUBTREE,
                    },
                    'call_fake': lambda *args, **kwargs: [
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                            'memberOf': [
                                b'cn=other,ou=Groups,dc=example,dc=com',
                            ],
                        }),
                    ],
                },
                {
                    'args': ['dc=example,dc=com'],
                    'kwargs': {
                        'filterstr': '(&(objectClass=group)(cn=other))',
                        'scope': ldap.SCOPE_SUBTREE,
                    },
                    'call_fake': lambda *args, **kwargs: [],
                },
            ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user@test.corp',
                                         password='test-pass')

        self.assertIsNone(user)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@test.corp.example.com',
            'test-pass')
        self.assertSpyCallCount(TestLDAPObject.search_s, 2)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=test,dc=corp,dc=example,dc=com',
            filterstr='(&(objectClass=user)(sAMAccountName=test-user))',
            scope=ldap.SCOPE_SUBTREE)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'dc=example,dc=com',
            filterstr='(&(objectClass=group)(cn=other))',
            scope=ldap.SCOPE_SUBTREE)

    def test_authenticate_with_invalid_credentials(self):
        """Testing ActiveDirectoryBackend.authenticate with invalid credentials
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpRaise(ldap.INVALID_CREDENTIALS()))
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                        }),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user',
                                         password='test-pass')

        self.assertIsNone(user)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@example.com',
            'test-pass')
        self.assertSpyNotCalled(TestLDAPObject.search_s)

    def test_authenticate_with_server_down(self):
        """Testing ActiveDirectoryBackend.authenticate with Server Down error
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpRaise(ldap.SERVER_DOWN()))
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                        }),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user',
                                         password='test-pass')

        self.assertIsNone(user)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@example.com',
            'test-pass')
        self.assertSpyNotCalled(TestLDAPObject.search_s)

    def test_authenticate_with_exception(self):
        """Testing ActiveDirectoryBackend.authenticate with unexpected
        exception
        """
        self.spy_on(TestLDAPObject.simple_bind_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpRaise(Exception('Kaboom!')))
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Test User,OU=MyOrg,DC=example,DC=com', {
                            'givenName': [b'Test'],
                            'sn': [b'User'],
                            'mail': [b'test-user@corp.example.com'],
                        }),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='test-user',
                                         password='test-pass')

        self.assertIsNone(user)

        self.assertSpyCalledWith(
            TestLDAPObject.simple_bind_s,
            'test-user@example.com',
            'test-pass')
        self.assertSpyNotCalled(TestLDAPObject.search_s)

    def test_find_domain_controllers_from_dns(self):
        """Testing ActiveDirectoryBackend.find_domain_controllers_from_dns"""
        self.spy_on(
            dns.resolver.query,
            op=kgb.SpyOpReturn(dns.resolver.Answer(
                qname=dns.name.from_text('_ldap._tcp.example.com'),
                rdtype=dns.rdatatype.SRV,
                rdclass=dns.rdataclass.IN,
                response=dns.message.from_text(
                    'id 12345\n'
                    'opcode QUERY\n'
                    'rcode NOERROR\n'
                    'flags QR RD RA\n'
                    ';QUESTION\n'
                    '_ldap._tcp.example.com. IN SRV\n'
                    ';ANSWER\n'
                    '_ldap._tcp.example.com. 299 IN SRV 10 1 389'
                    ' my-ad3.example.com.\n'
                    '_ldap._tcp.example.com. 299 IN SRV 5 1 390'
                    ' my-ad2.example.com.\n'
                    '_ldap._tcp.example.com. 299 IN SRV 5 5 391'
                    ' my-ad1.example.com.\n'
                    ';AUTHORITY\n'
                    ';ADDITIONAL\n'))))

        self.assertEqual(
            self.backend.find_domain_controllers_from_dns(),
            [
                (391, 'my-ad1.example.com'),
                (390, 'my-ad2.example.com'),
                (389, 'my-ad3.example.com'),
            ])

        self.assertSpyCalledWith(dns.resolver.query,
                                 '_ldap._tcp.example.com',
                                 rdtype='SRV')

    def test_find_domain_controllers_from_dns_with_custom_domain(self):
        """Testing ActiveDirectoryBackend.find_domain_controllers_from_dns
        with custom domain
        """
        self.spy_on(
            dns.resolver.query,
            op=kgb.SpyOpReturn(dns.resolver.Answer(
                qname=dns.name.from_text('_ldap._tcp.corp.example.com'),
                rdtype=dns.rdatatype.SRV,
                rdclass=dns.rdataclass.IN,
                response=dns.message.from_text(
                    'id 12345\n'
                    'opcode QUERY\n'
                    'rcode NOERROR\n'
                    'flags QR RD RA\n'
                    ';QUESTION\n'
                    '_ldap._tcp.corp.example.com. IN SRV\n'
                    ';ANSWER\n'
                    '_ldap._tcp.corp.example.com. 299 IN SRV 10 1 389'
                    ' my-ad3.example.com.\n'
                    '_ldap._tcp.corp.example.com. 299 IN SRV 5 1 390'
                    ' my-ad2.example.com.\n'
                    '_ldap._tcp.corp.example.com. 299 IN SRV 5 5 391'
                    ' my-ad1.example.com.\n'
                    ';AUTHORITY\n'
                    ';ADDITIONAL\n'))))

        self.assertEqual(
            self.backend.find_domain_controllers_from_dns('corp.example.com'),
            [
                (391, 'my-ad1.example.com'),
                (390, 'my-ad2.example.com'),
                (389, 'my-ad3.example.com'),
            ])

        self.assertSpyCalledWith(dns.resolver.query,
                                 '_ldap._tcp.corp.example.com',
                                 rdtype='SRV')

    def test_find_domain_controllers_from_dns_with_not_found(self):
        """Testing ActiveDirectoryBackend.find_domain_controllers_from_dns
        with domain not found
        """
        self.spy_on(dns.resolver.query,
                    op=kgb.SpyOpRaise(dns.resolver.NXDOMAIN()))

        self.assertEqual(self.backend.find_domain_controllers_from_dns(), [])

        self.assertSpyCalledWith(dns.resolver.query,
                                 '_ldap._tcp.example.com',
                                 rdtype='SRV')

    def test_find_domain_controllers_from_dns_with_error(self):
        """Testing ActiveDirectoryBackend.find_domain_controllers_from_dns
        with error
        """
        self.spy_on(dns.resolver.query,
                    op=kgb.SpyOpRaise(Exception('Kaboom!')))

        self.assertEqual(self.backend.find_domain_controllers_from_dns(), [])

        self.assertSpyCalledWith(dns.resolver.query,
                                 '_ldap._tcp.example.com',
                                 rdtype='SRV')

    @override_settings(AD_SEARCH_ROOT='ou=Test,dc=example,dc=com')
    def test_get_ldap_search_root_with_setting(self):
        """Testing ActiveDirectoryBackend.get_ldap_search_root with
        AD_SEARCH_ROOT
        """
        self.assertEqual(self.backend.get_ldap_search_root(),
                         'ou=Test,dc=example,dc=com')

    def test_get_ldap_search_root_with_defaults(self):
        """Testing ActiveDirectoryBackend.get_ldap_search_root with
        defaults
        """
        self.assertEqual(self.backend.get_ldap_search_root(),
                         'dc=example,dc=com')

    def test_get_ldap_search_root_with_custom_domain(self):
        """Testing ActiveDirectoryBackend.get_ldap_search_root with
        custom user_domain
        """
        self.assertEqual(self.backend.get_ldap_search_root('corp.example.com'),
                         'dc=corp,dc=example,dc=com')

    @override_settings(AD_OU_NAME='Test')
    def test_get_ldap_search_root_with_ad_ou_name(self):
        """Testing ActiveDirectoryBackend.get_ldap_search_root with
        AD_OU_NAME
        """
        self.assertEqual(self.backend.get_ldap_search_root(),
                         'ou=Test,dc=example,dc=com')

    def test_get_or_create_user_without_ad_user_data_and_with_user(self):
        """Testing ActiveDirectoryBackend.get_or_create_user without
        ad_user_data and with user in database
        """
        user = User.objects.create(username='test')

        self.assertEqual(self.backend.get_or_create_user('test'), user)

    def test_get_or_create_user_without_ad_user_data_and_without_user(self):
        """Testing ActiveDirectoryBackend.get_or_create_user without
        ad_user_data and with user not in database
        """
        self.assertIsNone(self.backend.get_or_create_user('test'))

    def test_get_ldap_connections(self):
        """Testing ActiveDirectoryBackend.get_ldap_connections"""
        connections = list(self.backend.get_ldap_connections('example.com'))

        self.assertEqual(len(connections), 2)
        self.assertEqual(connections[0][0], 'ldap://ad1.example.com:389')
        self.assertEqual(connections[1][0], 'ldap://ad2.example.com:123')

    @override_settings(AD_FIND_DC_FROM_DNS=True)
    def test_get_ldap_connections_from_dns(self):
        """Testing ActiveDirectoryBackend.get_ldap_connections with
        AD_FIND_DC_FROM_DNS=True
        """
        self.spy_on(
            dns.resolver.query,
            op=kgb.SpyOpReturn(dns.resolver.Answer(
                qname=dns.name.from_text('_ldap._tcp.example.com'),
                rdtype=dns.rdatatype.SRV,
                rdclass=dns.rdataclass.IN,
                response=dns.message.from_text(
                    'id 12345\n'
                    'opcode QUERY\n'
                    'rcode NOERROR\n'
                    'flags QR RD RA\n'
                    ';QUESTION\n'
                    '_ldap._tcp.example.com. IN SRV\n'
                    ';ANSWER\n'
                    '_ldap._tcp.example.com. 299 IN SRV 10 1 389'
                    ' my-ad3.example.com.\n'
                    '_ldap._tcp.example.com. 299 IN SRV 5 1 390'
                    ' my-ad2.example.com.\n'
                    '_ldap._tcp.example.com. 299 IN SRV 5 5 391'
                    ' my-ad1.example.com.\n'
                    ';AUTHORITY\n'
                    ';ADDITIONAL\n'))))

        connections = list(self.backend.get_ldap_connections('example.com'))

        self.assertEqual(len(connections), 3)
        self.assertEqual(connections[0][0], 'ldap://my-ad1.example.com:391')
        self.assertEqual(connections[1][0], 'ldap://my-ad2.example.com:390')
        self.assertEqual(connections[2][0], 'ldap://my-ad3.example.com:389')

    @override_settings(AD_FIND_DC_FROM_DNS=True)
    def test_get_ldap_connections_from_dns_with_not_found(self):
        """Testing ActiveDirectoryBackend.get_ldap_connections with
        AD_FIND_DC_FROM_DNS=True with domain not found
        """
        self.spy_on(dns.resolver.query,
                    op=kgb.SpyOpRaise(dns.resolver.NXDOMAIN()))

        self.assertEqual(
            list(self.backend.get_ldap_connections('example.com')),
            [])

    @override_settings(AD_FIND_DC_FROM_DNS=True)
    def test_get_ldap_connections_from_dns_with_error(self):
        """Testing ActiveDirectoryBackend.get_ldap_connections with
        AD_FIND_DC_FROM_DNS=True with error
        """
        self.spy_on(dns.resolver.query,
                    op=kgb.SpyOpRaise(Exception('Kaboom!')))

        self.assertEqual(
            list(self.backend.get_ldap_connections('example.com')),
            [])



"""Methods to help with building the review request rendering context."""

from __future__ import annotations

from typing import Any, Optional, TYPE_CHECKING, Union

from django.utils.translation import gettext as _
from django.template.defaultfilters import truncatechars
from djblets.siteconfig.models import SiteConfiguration
from housekeeping import deprecate_non_keyword_only_args
from typing_extensions import NotRequired, TypedDict

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.admin.server import build_server_url
from reviewboard.deprecation import (RemovedInReviewBoard90Warning,
                                     RemovedInReviewBoard10_0Warning)
from reviewboard.diffviewer.models import DiffSet
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from datetime import datetime

    from django.http import HttpRequest

    from reviewboard.reviews.models import (
        ReviewRequest,
        ReviewRequestDraft,
    )
    from reviewboard.reviews.models.review_request import \
        ReviewRequestCloseInfo
    from reviewboard.scmtools.models import Tool


class SerializedReviewRequestTab(TypedDict):
    """Serialized information about a tab on the review request page.

    Version Added:
        7.0
    """

    #: Whether this tab is the active tab.
    active: NotRequired[bool]

    #: The text to show on the tab label.
    text: str

    #: The URL to link to for the tab.
    url: str


class ReviewRequestContext(TypedDict):
    """Template context for rendering the review request.

    Version Changed:
        7.0.2:
        Added the following fields:

        * ``close_description``
        * ``close_description_rich_text``
        * ``close_timestamp``
        * ``draft``
        * ``force_view_user_draft``
        * ``review_request_details``
        * ``user_draft_exists``
        * ``viewing_user_draft``

    Version Added:
        7.0
    """

    #: The description attached to the most recent closing.
    #:
    #: Version Added:
    #:     7.0.2
    close_description: str

    #: Whether the close description is rich text.
    #:
    #: Version Added:
    #:     7.0.2
    close_description_rich_text: bool

    #: The timestamp of the most recent closing.
    #:
    #: Version Added:
    #:     7.0.2
    close_timestamp: Optional[datetime]

    #: The current draft, if present.
    #:
    #: Version Added:
    #:     7.0.2
    draft: Optional[ReviewRequestDraft]

    #: Whether to force viewing a draft owned by another user.
    #:
    #: Version Added:
    #:     7.0.2
    force_view_user_draft: bool

    #: Whether the review request is mutable by the current user.
    mutable_by_user: bool

    #: The review request object.
    review_request: ReviewRequest

    #: The current object to use for displaying the review request data.
    #:
    #: Version Added:
    #:     7.0.2
    review_request_details: Union[ReviewRequest, ReviewRequestDraft]

    #: The most recent review request visit info, if available.
    review_request_visit: NotRequired[ReviewRequestVisit]

    #: Whether the user can change the review request status.
    status_mutable_by_user: bool

    #: The SCMTool for the current review request, if it has a diff.
    scmtool: Optional[Tool]

    #: Global setting for whether to send e-mails for publish actions.
    send_email: bool

    #: The description to use for social media links to the review request.
    social_page_description: str

    #: The image URL to use for social media links.
    social_page_image_url: Optional[str]

    #: The title text to use for social media links.
    social_page_title: str

    #: The URL to use for social media links to the review request.
    social_page_url: str

    #: The tabs to show for the review request.
    tabs: list[SerializedReviewRequestTab]

    #: Whether a draft exists that is owned by another user.
    #:
    #: Version Added:
    #:     7.0.2
    user_draft_exists: bool

    #: Whether the user is viewing a draft owned by another user.
    #:
    #: Version Added:
    #:     7.0.2
    viewing_user_draft: bool


@deprecate_non_keyword_only_args(RemovedInReviewBoard90Warning)
def make_review_request_context(
    *,
    request: HttpRequest,
    review_request: ReviewRequest,
    extra_context: Optional[dict[str, Any]] = None,
    is_diff_view: bool = False,
    social_page_image_url: Optional[str] = None,
    social_page_title: str = '',
    review_request_details: Optional[
        Union[ReviewRequest, ReviewRequestDraft]] = None,
    close_info: Optional[ReviewRequestCloseInfo] = None,
    draft: Optional[ReviewRequestDraft] = None,
    force_view_user_draft: bool = False,
) -> ReviewRequestContext:
    """Return a dictionary for template contexts used for review requests.

    The dictionary will contain the common data that is used for all
    review request-related pages (the review request detail page, the diff
    viewer, and the screenshot pages).

    For convenience, extra data can be passed to this dictionary.

    Version Changed:
        7.0.2:
        * Deprecated the ``extra_context`` argument.
        * Added the ``review_request_details``, ``close_info``, ``draft`` and
          ``force_view_user_draft`` arguments.

    Version Changed:
        7.0:
        Added ``social_page_image_url`` and ``social_page_title`` arguments.

    Args:
        request (django.http.HttpRequest):
            The HTTP request.

        review_request (reviewboard.reviews.models.ReviewRequest):
            The review request.

        extra_context (dict, optional):
            Extra information to include in the context.

        is_diff_view (bool, optional):
            Whether the user is viewing a diff.

        social_page_image_url (str, optional):
            The image URL to include for social media thumbnails.

            Version Added:
                7.0

        social_page_title (str, optional):
            The page title to include for social media thumbnails.

            Version Added:
                7.0

        review_request_details (reviewboard.reviews.models.ReviewRequest or
                                reviewboard.reviews.models.
                                ReviewRequestDraft):
            The object to show review request content from.

            Version Added:
                7.0

        close_info (reviewboard.reviews.models.review_request.
                    ReviewRequestCloseInfo, optional):
            Close information about the review request.

            Version Added:
                7.0.2

        draft (reviewboard.reviews.models.ReviewRequestDraft, optional):
            The review request draft.

            Version Added:
                7.0.2

        force_view_user_draft (bool, optional):
            Whether to force viewing another user's draft.

            This is used when a user with privilege to view other users' draft
            data is viewing something which only exists in the draft (such as
            an unpublished review request or a file attachment added in a new
            draft).

            Version Added:
                7.0.2

    Returns:
        ReviewRequestContext:
        The context for rendering review request templates.
    """
    if review_request.repository:
        scmtool = review_request.repository.get_scmtool()
    else:
        scmtool = None

    tabs: list[SerializedReviewRequestTab] = [
        {
            'text': _('Reviews'),
            'url': review_request.get_absolute_url(),
        },
    ]

    if draft is None:
        draft = review_request.get_draft(request.user)

    # If we have an accessible draft, and the viewing user is not the owner of
    # the review request, we normally do not want to show the draft data. If
    # the page is accessed via ?view-draft=1, we'll let the admin-like user see
    # and manipulate the draft data.
    user_draft_exists: bool = False
    viewing_user_draft: bool = False

    # If the user viewing this review request is not the submitter, but has
    # access to the draft, we want to let them choose to either view it as if
    # they were a regular user, or view the draft data. In the case of review
    # requests which are not yet public, we unconditionally show the draft
    # data.
    if draft and request.user != review_request.submitter:
        user_draft_exists = True

        if (force_view_user_draft or
            should_view_draft(request=request,
                              review_request=review_request,
                              draft=draft)):
            viewing_user_draft = True
        else:
            draft = None

        if not review_request.public:
            force_view_user_draft = True

    if ((draft and draft.diffset_id) or
        (hasattr(review_request, '_diffsets') and
         len(review_request._diffsets) > 0)):
        has_diffs = True
    else:
        # We actually have to do a query
        has_diffs = DiffSet.objects.filter(
            history__pk=review_request.diffset_history_id).exists()

    if has_diffs:
        tabs.append({
            'active': is_diff_view,
            'text': _('Diff'),
            'url': (
                local_site_reverse(
                    'view-diff',
                    args=[review_request.display_id],
                    local_site=review_request.local_site) +
                '#index_header'),
        })

    siteconfig = SiteConfiguration.objects.get_current()

    if review_request_details is None:
        review_request_details = draft or review_request

    description = review_request_details.description
    assert description is not None
    social_page_description = truncatechars(
        description.replace('\n', ' '), 300)

    if close_info is None:
        close_info = review_request.get_close_info()

    context: ReviewRequestContext = {
        'close_description': close_info['close_description'],
        'close_description_rich_text': close_info['is_rich_text'],
        'close_timestamp': close_info['timestamp'],
        'draft': draft,
        'force_view_user_draft': force_view_user_draft,
        'mutable_by_user': review_request.is_mutable_by(request.user),
        'status_mutable_by_user':
            review_request.is_status_mutable_by(request.user),
        'review_request': review_request,
        'review_request_details': review_request_details,
        'scmtool': scmtool,
        'send_email': siteconfig.get('mail_send_review_mail'),
        'tabs': tabs,
        'social_page_description': social_page_description,
        'social_page_image_url': social_page_image_url,
        'social_page_title': social_page_title,
        'social_page_url': build_server_url(request.path, request=request),
        'user_draft_exists': user_draft_exists,
        'viewing_user_draft': viewing_user_draft,
    }

    if extra_context:
        RemovedInReviewBoard10_0Warning.warn(
            'The extra_context argument to make_review_request_context is '
            'deprecated and will be removed in Review Board 10.')

        context.update(extra_context)

    if ('review_request_visit' not in context and
        request.user.is_authenticated):
        # The main review request view will already have populated this, but
        # other related views (like the diffviewer) don't.
        context['review_request_visit'] = \
            ReviewRequestVisit.objects.get_or_create(
                user=request.user,
                review_request=review_request)[0]

    return context


def should_view_draft(
    *,
    request: HttpRequest,
    review_request: ReviewRequest,
    draft: Optional[ReviewRequestDraft],
) -> bool:
    """Return whether the requesting user should view the draft.

    Args:
        request (django.http.HttpRequest):
            The HTTP request.

        review_request (reviewboard.reviews.models.ReviewRequest):
            The current review request.

        draft (reviewboard.reviews.models.ReviewRequestDraft):
            The current review request draft.

    Returns:
        bool:
        ``True`` if the user should view the draft data.
    """
    if draft:
        # The owner of the review request should always see the draft data.
        if review_request.submitter == request.user:
            return True

        # Review requests that have not yet been published only have draft
        # data.
        if not review_request.public:
            return True

        # If the user wants to view the draft, show it.
        if request.GET.get('view-draft', False) == '1':
            return True

    return False



"""Unit tests for reviewboard.certs.errors.CertificateVerificationError.

Version Added:
    6.0
"""

from __future__ import annotations

from datetime import datetime, timezone

from reviewboard.certs.cert import Certificate, CertificateFingerprints
from reviewboard.certs.errors import (CertificateVerificationError,
                                      CertificateVerificationFailureCode)
from reviewboard.certs.tests.testcases import (CertificateTestCase,
                                               TEST_SHA1,
                                               TEST_SHA256)


_TEST_CERT = Certificate(
    hostname='example.com',
    port=443,
    subject='Subject',
    issuer='Issuer',
    fingerprints=CertificateFingerprints(sha1=TEST_SHA1,
                                         sha256=TEST_SHA256),
    valid_from=datetime(2023, 7, 14, 7, 50, 30, tzinfo=timezone.utc),
    valid_through=datetime(2024, 7, 13, 7, 50, 30,
                           tzinfo=timezone.utc))

_TEST_CERT_DETAILS = (
    'hostname="example.com", port=443, issuer="Issuer", fingerprints='
    'SHA1=F2:35:0F:BB:34:40:84:78:8B:20:1D:40:B1:4A:17:0C:DE:36:2F:D5; '
    'SHA256=79:19:70:AE:A6:1B:EB:BC:35:7C:B8:54:B1:6A:AD:79:FF:F7:28:'
    '69:02:5E:C3:6F:B3:C2:B4:FD:84:66:DF:8F'
)


class CertificateVerificationErrorTests(CertificateTestCase):
    """Unit tests for CertificateVerificationError.

    Version Added:
        6.0
    """

    def test_init_with_expired(self) -> None:
        """Testing CertificateVerificationError with code=EXPIRED"""
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.EXPIRED)

        self.assertEqual(
            str(error),
            'The SSL certificate provided by the server has expired and can '
            'no longer be used.')
        self.assertEqual(error.generic_msg, str(error))
        self.assertIsNone(error.certificate)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.EXPIRED)

    def test_init_with_expired_and_cert(self) -> None:
        """Testing CertificateVerificationError with code=EXPIRED and
        certificate
        """
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.EXPIRED,
            certificate=_TEST_CERT)

        self.assertEqual(
            str(error),
            f'The SSL certificate provided by example.com has expired and '
            f'can no longer be used. Certificate details: '
            f'{_TEST_CERT_DETAILS}')
        self.assertEqual(
            error.generic_msg,
            'The SSL certificate provided by example.com has expired and '
            'can no longer be used.')
        self.assertEqual(error.certificate, _TEST_CERT)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.EXPIRED)

    def test_init_with_not_yet_valid(self) -> None:
        """Testing CertificateVerificationError with code=NOT_YET_VALID"""
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.NOT_YET_VALID)

        self.assertEqual(
            str(error),
            'The SSL certificate provided by the server is not yet valid '
            'and cannot be used.')
        self.assertEqual(error.generic_msg, str(error))
        self.assertIsNone(error.certificate)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.NOT_YET_VALID)

    def test_init_with_not_yet_valid_and_cert(self) -> None:
        """Testing CertificateVerificationError with code=NOT_YET_VALID and
        certificate
        """
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.NOT_YET_VALID,
            certificate=_TEST_CERT)

        self.assertEqual(
            str(error),
            f'The SSL certificate provided by example.com is not yet valid '
            f'and cannot be used. Certificate details: {_TEST_CERT_DETAILS}')
        self.assertEqual(
            error.generic_msg,
            'The SSL certificate provided by example.com is not yet valid '
            'and cannot be used.')
        self.assertEqual(error.certificate, _TEST_CERT)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.NOT_YET_VALID)

    def test_init_with_hostname_mismatch(self) -> None:
        """Testing CertificateVerificationError with code=HOSTNAME_MISMATCH"""
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.HOSTNAME_MISMATCH)

        self.assertEqual(
            str(error),
            'The SSL certificate provided by the server does not match its '
            'hostname and may not be safe.')
        self.assertEqual(error.generic_msg, str(error))
        self.assertIsNone(error.certificate)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.HOSTNAME_MISMATCH)

    def test_init_with_hostname_mismatch_and_cert(self) -> None:
        """Testing CertificateVerificationError with code=HOSTNAME_MISMATCH
        and certificate
        """
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.HOSTNAME_MISMATCH,
            certificate=_TEST_CERT)

        self.assertEqual(
            str(error),
            f'The SSL certificate provided by example.com does not match its '
            f'hostname and may not be safe. Certificate details: '
            f'{_TEST_CERT_DETAILS}')
        self.assertEqual(
            error.generic_msg,
            'The SSL certificate provided by example.com does not match its '
            'hostname and may not be safe.')
        self.assertEqual(error.certificate, _TEST_CERT)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.HOSTNAME_MISMATCH)

    def test_init_with_not_trusted(self) -> None:
        """Testing CertificateVerificationError with code=NOT_TRUSTED"""
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.NOT_TRUSTED)

        self.assertEqual(
            str(error),
            'The SSL certificate provided by the server has not been signed '
            'by a trusted Certificate Authority and may not be safe. The '
            'certificate needs to be verified in Review Board before the '
            'server can be accessed.')
        self.assertEqual(error.generic_msg, str(error))
        self.assertIsNone(error.certificate)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.NOT_TRUSTED)

    def test_init_with_not_trusted_and_cert(self) -> None:
        """Testing CertificateVerificationError with code=NOT_TRUSTED and
        certificate
        """
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.NOT_TRUSTED,
            certificate=_TEST_CERT)

        self.assertEqual(
            str(error),
            f'The SSL certificate provided by example.com has not been signed '
            f'by a trusted Certificate Authority and may not be safe. The '
            f'certificate needs to be verified in Review Board before the '
            f'server can be accessed. Certificate details: '
            f'{_TEST_CERT_DETAILS}')
        self.assertEqual(
            error.generic_msg,
            'The SSL certificate provided by example.com has not been signed '
            'by a trusted Certificate Authority and may not be safe. The '
            'certificate needs to be verified in Review Board before the '
            'server can be accessed.')
        self.assertEqual(error.certificate, _TEST_CERT)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.NOT_TRUSTED)

    def test_init_with_other(self) -> None:
        """Testing CertificateVerificationError with code=OTHER"""
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.OTHER)

        self.assertEqual(
            str(error),
            'The SSL certificate provided by the server could not be '
            'verified and may not be safe. The certificate must be valid '
            'and verified in Review Board before the server can be '
            'accessed.')
        self.assertEqual(error.generic_msg, str(error))
        self.assertIsNone(error.certificate)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.OTHER)

    def test_init_with_other_and_cert(self) -> None:
        """Testing CertificateVerificationError with code=OTHER and
        certificate
        """
        error = CertificateVerificationError(
            code=CertificateVerificationFailureCode.OTHER,
            certificate=_TEST_CERT)

        self.assertEqual(
            str(error),
            f'The SSL certificate provided by example.com could not be '
            f'verified and may not be safe. The certificate must be valid '
            f'and verified in Review Board before the server can be '
            f'accessed. Certificate details: {_TEST_CERT_DETAILS}')
        self.assertEqual(
            error.generic_msg,
            'The SSL certificate provided by example.com could not be '
            'verified and may not be safe. The certificate must be valid '
            'and verified in Review Board before the server can be '
            'accessed.')
        self.assertEqual(error.certificate, _TEST_CERT)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.OTHER)

    def test_init_with_custom_message(self) -> None:
        """Testing CertificateVerificationError with custom message"""
        error = CertificateVerificationError(
            'Something is horribly wrong with the certificate (error '
            '%(code)s)!',
            code=CertificateVerificationFailureCode.NOT_TRUSTED)

        self.assertEqual(
            str(error),
            'Something is horribly wrong with the certificate (error '
            'NOT_TRUSTED)!')
        self.assertEqual(error.generic_msg, str(error))
        self.assertIsNone(error.certificate)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.NOT_TRUSTED)

    def test_init_with_custom_message_and_cert(self) -> None:
        """Testing CertificateVerificationError with code=NOT_TRUSTED and
        certificate
        """
        error = CertificateVerificationError(
            'Something is horribly wrong with the %(hostname)s certificate '
            '(error %(code)s)!',
            code=CertificateVerificationFailureCode.NOT_TRUSTED,
            certificate=_TEST_CERT)

        self.assertEqual(
            str(error),
            f'Something is horribly wrong with the example.com certificate '
            f'(error NOT_TRUSTED)! Certificate details: {_TEST_CERT_DETAILS}')
        self.assertEqual(
            error.generic_msg,
            'Something is horribly wrong with the example.com certificate '
            '(error NOT_TRUSTED)!')
        self.assertEqual(error.certificate, _TEST_CERT)
        self.assertEqual(error.code,
                         CertificateVerificationFailureCode.NOT_TRUSTED)



"""Base classes for datagrid unit tests.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Optional

from django.contrib.auth.models import User
from django.db.models import QuerySet
from django.test.client import RequestFactory
from djblets.datagrid.grids import Column, DataGrid
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.models import LocalSiteProfile, Profile
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class BaseViewTestCase(TestCase):
    """Base class for tests of dashboard views."""

    #: The base URL to the datagrid being tested.
    #:
    #: Type:
    #:     str
    #:
    #: Version Added:
    #:     5.0.7
    datagrid_url: str = ''

    def setUp(self) -> None:
        """Set up the test case.

        This will set some initial configuration defaults for access control.

        It will temporarily patch :py:meth:`django.db.models.QuerySet.
        __eq__` to help compare with nested queries. This is a temporary
        issue, and this function will soon be removed.
        """
        super().setUp()

        self.siteconfig = SiteConfiguration.objects.get_current()
        self._old_auth_require_sitewide_login = \
            self.siteconfig.get('auth_require_sitewide_login')
        self.siteconfig.set('auth_require_sitewide_login', False)
        self.siteconfig.save()

        # This is a very temporary hack to work around some assert_queries
        # comparisons that fail due to our improper use of a nested query.
        # It will be removed as soon as this issue is fixed.
        self._old_queryset_eq = QuerySet.__eq__

        setattr(QuerySet, '__eq__',
                lambda _self, other: repr(_self) == repr(other))

    def tearDown(self):
        """Tear down test state.

        This will restore the configuration and reset
        :py:meth:`django.db.models.QuerySet.__eq__` to defaults.
        """
        self.siteconfig = SiteConfiguration.objects.get_current()
        self.siteconfig.set('auth_require_sitewide_login',
                            self._old_auth_require_sitewide_login)
        self.siteconfig.save()

        setattr(QuerySet, '__eq__', self._old_queryset_eq)

        super().tearDown()

    def get_datagrid_url(
        self,
        *,
        local_site: Optional[LocalSite],
    ) -> str:
        """Return the URL to the datagrid page.

        This will take a provided Local Site into account for the URL.

        Version Added:
            5.0.7

        Args:
            local_site (reviewboard.site.models.LocalSite):
                The Local Site to access, or ``None`` if accessing the
                global site.

        Returns:
            str:
            The datagrid page URL.
        """
        url = self.datagrid_url

        if local_site is not None:
            url = '/s/%s%s' % (local_site.name, url)

        return url

    def _prefetch_cached(
        self,
        *,
        local_site: Optional[LocalSite] = None,
        user: Optional[User] = None,
    ) -> None:
        """Pre-fetch cacheable statistics and data.

        Version Added:
            5.0

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site being used for the test.
        """
        SiteConfiguration.objects.get_current()

        if local_site is not None:
            LocalSite.objects.get_local_site_acl_stats(local_site)

        for temp_user in User.objects.all():
            temp_user.get_local_site_stats()

        if user is not None:
            try:
                profile = user.get_profile(create_if_missing=False)
                profile.has_starred_review_groups(local_site=local_site)
                profile.has_starred_review_requests(local_site=local_site)
            except Profile.DoesNotExist:
                pass

            try:
                user.get_site_profile(local_site=local_site,
                                      create_if_missing=False)
            except LocalSiteProfile.DoesNotExist:
                pass

    def _get_context_var(self, response, varname):
        """Return a variable from the view context."""
        return response.context.get(varname)


class BaseColumnTestCase(TestCase):
    """Base class for defining a column unit test."""

    #: An instance of the column to use on the datagrid.
    #:
    #: Type:
    #:     djblets.datagrid.columns.Column
    column: Column

    fixtures = ['test_users']

    def setUp(self) -> None:
        super().setUp()

        assert getattr(self, 'column', None) is not None

        class TestDataGrid(DataGrid):
            column = self.column

        request_factory = RequestFactory()
        self.request = request_factory.get('/')
        self.request.user = User.objects.get(username='doc')

        self.grid = TestDataGrid(request=self.request)
        self.stateful_column = self.grid.get_stateful_column(self.column)



"""Extension hooks for managing actions."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint
from djblets.registries.errors import ItemLookupError

from reviewboard.actions import (AttachmentPoint,
                                 BaseAction,
                                 BaseMenuAction,
                                 actions_registry)
from reviewboard.actions.renderers import DefaultActionRenderer
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.urls import (diffviewer_url_names,
                              main_review_request_url_name)

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from django.template import Context


logger = logging.getLogger(__name__)


class ActionHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for injecting clickable actions into the UI.

    Actions are displayed either on the action bar of each review request or in
    the page header.

    The provided ``actions`` parameter must be a list of actions. These are
    subclasses of :py:class:`reviewboard.actions.base.BaseAction`.
    """

    #: The actions registered by this hook.
    actions: Sequence[BaseAction]

    def initialize(
        self,
        actions: (Sequence[BaseAction] | None) = None,
        *args,
        **kwargs,
    ) -> None:
        """Initialize this action hook.

        Args:
            actions (list of reviewboard.actions.base.BaseAction, optional):
                The list of actions to be added.

            *args (tuple):
                Extra positional arguments.

            **kwargs (dict):
                Extra keyword arguments.
        """
        self.actions = actions or []

        for action in self.actions:
            actions_registry.register(action)

    def shutdown(self) -> None:
        """Shut down the hook."""
        super().shutdown()

        for action in self.actions:
            try:
                actions_registry.unregister(action)
            except ItemLookupError:
                pass

    def get_actions(
        self,
        context: Context,
    ) -> Sequence[BaseAction]:
        """Return the list of action information for this action hook.

        Args:
            context (django.template.Context):
                The collection of key-value pairs available in the template.

        Returns:
            list: The list of action information for this action hook.
        """
        return self.actions


class _DictAction(BaseAction):
    """An action for legacy dictionary-based actions.

    For backwards compatibility, review request actions may also be supplied as
    :py:class:`ActionHook`-style dictionaries. This helper class is used by
    :py:meth:`convert_action` to convert these types of dictionaries into
    instances of :py:class:`reviewboard.actions.base.BaseAction`.
    """

    default_renderer_cls = DefaultActionRenderer
    _ignore_action_deprecations = True

    def __init__(self, action_dict, applies_to, attachment):
        """Initialize this action.

        Args:
            action_dict (dict):
                A dictionary representing this action, as specified by the
                :py:class:`ActionHook` class.

            applies_to (callable):
                The list of URLs to apply the action to.

            attachment (reviewboard.actions.base.AttachmentPoint):
                The attachment point for the action.
        """
        if not isinstance(action_dict, dict):
            raise ValueError('Action definitions must be dictionaries.')

        self.label = action_dict['label']

        if 'id' in action_dict:
            self.action_id = action_dict['id']
        else:
            self.action_id = re.sub(
                r'[^a-zA-Z0-9]+',
                '-',
                self.label.lower())

        self.attachment = attachment

        super().__init__()

        self.url = action_dict['url']
        self.apply_to = applies_to

        if 'image' in action_dict:
            self.image = action_dict['image']
            self.image_width = action_dict['image_width']
            self.image_height = action_dict['image_height']

    def get_dom_element_id(self) -> str:
        """Return the ID used for the DOM element for this action.

        Returns:
            str:
            The ID used for the element.
        """
        # We use this instead of the new style (action-*) in order to maintain
        # compatibility with the old implementation.
        return '%s-action' % self.action_id


class _DictMenuAction(BaseMenuAction):
    """A menu action for legacy dictionary-based menu actions.

    For backwards compatibility, review request actions may also be supplied as
    :py:class:`ReviewRequestDropdownActionHook`-style dictionaries. This helper
    class is used by :py:meth:`convert_action` to convert these types of
    dictionaries into instances of
    :py:class:`reviewboard.actions.base.BaseMenuAction`.
    """

    _ignore_action_deprecations = True

    def __init__(self, action_dict, applies_to, attachment):
        """Initialize this action.

        Args:
            action_dict (dict):
                A dictionary representing this menu action, as specified by the
                :py:class:`ReviewRequestDropdownActionHook` class.

            applies_to (list of str):
                The list of URLs to apply the action to.

            attachment (reviewboard.actions.base.AttachmentPoint):
                The attachment point for the action.
        """
        if not isinstance(action_dict, dict):
            raise ValueError('Action definitions must be dictionaries.')

        self.label = action_dict['label']

        if 'id' in action_dict:
            self.action_id = action_dict['id']
        else:
            self.action_id = re.sub(
                r'[^a-zA-Z0-9]*',
                '-',
                self.label.lower())

        self.attachment = attachment

        super().__init__()

        self.apply_to = applies_to

    def get_dom_element_id(self) -> str:
        """Return the ID used for the DOM element for this action.

        Returns:
            str:
            The ID used for the element.
        """
        return f'action-{self.action_id}'


class BaseReviewRequestActionHook(ActionHook, metaclass=ExtensionHookPoint):
    """A base hook for adding review request actions to the action bar.

    Review request actions are displayed on the action bar (alongside default
    actions such as :guilabel:`Download Diff` and :guilabel:`Ship It!`) of each
    review request.
    """

    attachment = AttachmentPoint.REVIEW_REQUEST

    def initialize(self, actions=[], apply_to=None, *args, **kwargs):
        """Initialize this action hook.

        Args:
            actions (list of dict, optional):
                The list of actions to be added.

            apply_to (list of unicode, optional):
                The list of URL names that this action hook will apply to.

            *args (tuple):
                Extra positional arguments.

            **kwargs (dict):
                Extra keyword arguments.

        Raises:
            KeyError:
                Some dictionary is not an :py:class:`ActionHook`-style
                dictionary.
        """
        self.applies_to = apply_to
        super().initialize(
            actions=[
                self.convert_action(action_dict)
                for action_dict in actions
            ], *args, **kwargs)

    def convert_action(self, action_dict):
        """Convert the given dictionary to a review request action instance.

        Args:
            action_dict (dict):
                A dictionary representing a review request action, as specified
                by the :py:class:`ActionHook` class.

        Returns:
            reviewboard.actions.base.BaseAction:
            The corresponding review request action instance.

        Raises:
            KeyError:
                The given dictionary does not have the correct fields.

            ValueError:
                A given action was not a dictionary.
        """
        if not isinstance(action_dict, dict):
            raise ValueError('Action definitions must be dictionaries.')

        for key in ('label', 'url'):
            if key not in action_dict:
                raise KeyError('Action dictionaries require a %s key'
                               % repr(key))

        return _DictAction(action_dict, self.applies_to, self.attachment)


class ReviewRequestActionHook(BaseReviewRequestActionHook,
                              metaclass=ExtensionHookPoint):
    """A hook for adding review request actions to review request pages.

    By default, actions that are passed into this hook will only be displayed
    on review request pages and not on any file attachment pages or diff
    viewer pages.

    This hook is deprecated. New extensions should use
    :py:class:`reviewboard.extensions.hooks.ActionHook`.

    The provided ``actions`` parameter must be a list of actions. Each action
    must be a :py:class:`dict` with the following keys:

    ``id`` (:py:class:`str`, optional):
        The ID to use for the action.

    ``label`` (:py:class:`str`):
        The label to use for the action.

    ``url`` (:py:class:`str`):
        The URL to link the action to.

        If you want to use JavaScript to handle the action, this should be
        ``"#"`` and you should attach a handler to the element specified by
        the "#action-<id>" selector.
    """

    def initialize(self, actions=[], apply_to=None):
        """Initialize this action hook.

        Args:
            actions (list of dict, optional):
                The list of actions to be added.

            apply_to (list of unicode, optional):
                The list of URL names that this action hook will apply to.
                By default, this will apply to the main review request page
                only.

        Raises:
            KeyError:
                Some dictionary is not an :py:class:`ActionHook`-style
                dictionary.
        """
        if apply_to is None:
            apply_to = [main_review_request_url_name]

        super().initialize(actions=actions, apply_to=apply_to)

        RemovedInReviewBoard90Warning.warn(
            'ReviewRequestActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension "%s" will need to be updated to '
            'derive actions from reviewboard.actions.BaseAction and use '
            'ActionHook.'
            % self.extension.id)


class ReviewRequestDropdownActionHook(ActionHook,
                                      metaclass=ExtensionHookPoint):
    """A hook for adding dropdown menu actions to review request pages.

    This hook is deprecated. New extensions should use
    :py:class:`reviewboard.extensions.hooks.ActionHook`.

    The provided ``actions`` parameter must be a list of actions. Each action
    must be a :py:class:`dict` with the following keys:

    ``id`` (:py:class:`str`, optional):
        The ID to use for the action.

    ``label`` (:py:class:`str`):
        The label to use for the action.

    ''items`` (:py:class:`list`):
        A list of items, each of which is a :py:class:`dict` which follows the
        conventions for :py:class:`ReviewRequestActionHook`.

    Example:
        .. code-block:: python

           actions = [{
               'id': 'sample-menu-action',
               'label': 'Sample Menu',
               'items': [
                   {
                       'id': 'first-item-action',
                       'label': 'Item 1',
                       'url': '#',
                   },
                   {
                       'label': 'Item 2',
                       'url': '#',
                   },
               ],
           }]
    """

    attachment = AttachmentPoint.REVIEW_REQUEST

    def initialize(self, actions=[], apply_to=None):
        """Initialize this action hook.

        Args:
            actions (list of dict, optional):
                The list of actions to be added.

            apply_to (list of unicode, optional):
                The list of URL names that this action hook will apply to.
                By default, this will apply to the main review request page
                only.

        Raises:
            KeyError:
                Some dictionary is not an :py:class:`ActionHook`-style
                dictionary.
        """
        converted_actions = []

        if actions is None:
            actions = []

        for action_dict in actions:
            action = _DictMenuAction(action_dict, apply_to, self.attachment)
            converted_actions.append(action)

            for child in action_dict['items']:
                child_action = _DictAction(child, apply_to, self.attachment)
                child_action.parent_id = action.action_id

                converted_actions.append(child_action)

        super().initialize(actions=converted_actions)

        RemovedInReviewBoard90Warning.warn(
            'ReviewRequestDropdownActionHook is deprecated and will be '
            'removed in Review Board 9. Your extension "%s" will need to be '
            'updated to derive actions from reviewboard.actions.BaseAction '
            'and use ActionHook.'
            % self.extension.id)


class DiffViewerActionHook(BaseReviewRequestActionHook,
                           metaclass=ExtensionHookPoint):
    """A hook for adding review request actions to diff viewer pages.

    By default, actions that are passed into this hook will only be displayed
    on diff viewer pages and not on any review request pages or file attachment
    pages.

    This hook is deprecated. New extensions should use
    :py:class:`reviewboard.extensions.hooks.ActionHook`.
    """

    def initialize(self, actions=[], apply_to=diffviewer_url_names):
        """Initialize this action hook.

        Args:
            actions (list of dict, optional):
                The list of actions to be added.

            apply_to (list of unicode, optional):
                The list of URL names that this action hook will apply to.

        Raises:
            KeyError:
                Some dictionary is not an :py:class:`ActionHook`-style
                dictionary.
        """
        super().initialize(
            actions,
            apply_to=apply_to or diffviewer_url_names)

        RemovedInReviewBoard90Warning.warn(
            'DiffViewerActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension "%s" will need to be updated to '
            'derive actions from reviewboard.actions.BaseAction and use '
            'ActionHook.'
            % self.extension.id)


class HeaderActionHook(BaseReviewRequestActionHook,
                       metaclass=ExtensionHookPoint):
    """A hook for adding actions to the page header.

    This hook is deprecated. New extensions should use
    :py:class:`reviewboard.extensions.hooks.ActionHook`.

    The provided ``actions`` parameter must be a list of actions. Each action
    must be a :py:class:`dict` with the following keys:

    ``id`` (:py:class:`str`, optional):
        The ID to use for the action.

    ``label`` (:py:class:`str`):
        The label to use for the action.

    ``url`` (:py:class:`str`):
        The URL to link the action to.

        If you want to use JavaScript to handle the action, this should be
        ``"#"`` and you should attach a handler to the element specified by
        the "#action-<id>" selector.

    ``image`` (:py:class:`str`):
        The URL to an image to display next to the label.

    ``image_width`` (:py:class:`int`):
        The width of the image, if present.

    ``image_height`` (:py:class:`int`):
        The height of the image, if present.
    """

    attachment = AttachmentPoint.HEADER

    def __init__(self, *args, **kwargs):
        """Initialize the hook.

        Args:
            *args (tuple):
                Positional arguments to pass through to the superclass.

            **kwargs (dict):
                Keyword arguments to pass through to the superclass.
        """
        super().__init__(*args, **kwargs)
        RemovedInReviewBoard90Warning.warn(
            'HeaderActionHook is deprecated and will be removed in Review '
            'Board 9. Your extension "%s" will need to be updated to derive '
            'actions from reviewboard.actions.BaseAction and use ActionHook.'
            % self.extension.id)

    def convert_action(self, action_dict):
        """Convert the given dictionary to na action instance.

        Args:
            action_dict (dict):
                A dictionary representing a header action.

        Returns:
            BaseReviewRequestMenuAction:
            The corresponding review request menu action instance.
        """
        action = super().convert_action(action_dict)

        if 'image' in action_dict:
            action.template_name = 'extensions/header_action_with_image.html'

        return action


class HeaderDropdownActionHook(ActionHook,
                               metaclass=ExtensionHookPoint):
    """A hook for adding dropdown menu actions to the page header.

    This hook is deprecated. New extensions should use
    :py:class:`reviewboard.extensions.hooks.ActionHook`.

    The provided ``actions`` parameter must be a list of actions. Each action
    must be a :py:class:`dict` with the following keys:

    ``id`` (:py:class:`str`, optional):
        The ID to use for the action.

    ``label`` (:py:class:`str`):
        The label to use for the action.

    ``items`` (:py:class:`list`):
        A list of items, each of which is a :py:class:`dict` which follows the
        conventions for :py:class:`HeaderActionHook`.
    """

    attachment = AttachmentPoint.HEADER

    def initialize(self, actions=[], *args, **kwargs):
        """Initialize the hook.

        Args:
            actions (list of dict, optional):
                The list of actions to be added.

            *args (tuple):
                Positional arguments to pass through to the superclass.

            **kwargs (dict):
                Keyword arguments to pass through to the superclass.
        """
        converted_actions = []

        if actions is None:
            actions = []

        for action_dict in actions:
            action = _DictMenuAction(action_dict, None, self.attachment)
            converted_actions.append(action)

            for child in action_dict['items']:
                child_action = _DictAction(child, None, self.attachment)
                child_action.parent_id = action.action_id

                converted_actions.append(child_action)

        super().initialize(actions=converted_actions)

        RemovedInReviewBoard90Warning.warn(
            'HeaderDropdownActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension "%s" will need to be updated to '
            'derive actions from reviewboard.actions.BaseAction and use '
            'ActionHook.'
            % self.extension.id)


class HideActionHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for hiding built-in actions.

    Extensions may want to replace built-in functionality with their own custom
    versions, or disable some things entirely (such as the quick ship-it
    button).
    """

    #: The list of action IDs hidden by this hook.
    hidden_action_ids: set[str]

    def initialize(
        self,
        action_ids: Iterable[str],
        *args,
        **kwargs,
    ) -> None:
        """Initialize the hook.

        Args:
            action_ids (list of str):
                The list of action IDs to hide.

            *args (tuple):
                Extra positional arguments.

            **kwargs (dict):
                Extra keyword arguments.
        """
        self.hidden_action_ids = set(action_ids)



"""HTTP handler support for SSL/TLS certificate verification.

Version Added:
    8.1
"""

from __future__ import annotations

import logging
import ssl
from typing import TYPE_CHECKING
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import (
    HTTPSHandler,
    Request,
    build_opener,
)

from typelets.symbols import UNSET

from reviewboard.certs.cert import Certificate
from reviewboard.certs.errors import (CertificateVerificationError,
                                      CertificateVerificationFailureCode)
from reviewboard.certs.manager import cert_manager as default_cert_manager

if TYPE_CHECKING:
    from http.client import HTTPResponse
    from urllib.request import _DataType, _UrlopenRet

    from typelets.funcs import KwargsDict
    from typelets.symbols import Unsettable

    from reviewboard.certs.manager import CertificateManager
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


class CertificateVerificationHTTPSHandler(HTTPSHandler):
    """An urlopen handler for CertificateManager-backed HTTPS connections.

    This handler specializes :py:class:`~urllib.request.HTTPSHandler`,
    using the Certificate Manager to provide an SSL context and converting
    any SSL/TLS errors into a
    :py:class:`~reviewboard.certs.errors.CertificateVerificationError` where
    possible.

    Version Added:
        8.1
    """

    ######################
    # Instance variables #
    ######################

    #: The associated Certificate Manager for the handler.
    _cert_manager: CertificateManager

    #: Optional PEM-formatted certificate data to use for verification.
    _extra_cert_data: str | None

    #: The Local Site used for certificate lookup.
    _local_site: LocalSite | None

    #: Whether we're checking the hostname against the certificate.
    #:
    #: We're storing this under a namespace, because the handler will
    #: normally set :py:attr:`_check_hostname` and trigger a deprecation
    #: error for that argument. We're controlling this ourselves and
    #: don't want to risk informing the parent of this.
    _rb_check_hostname: bool

    def __init__(
        self,
        *,
        local_site: (LocalSite | None),
        cert_manager: (CertificateManager | None) = None,
        check_hostname: bool = True,
        extra_cert_data: (str | None) = None,
    ) -> None:
        """Initialize the handler.

        Args:
            local_site (reviewboard.site.models.LocalSite):
                The Local Site the certificates would be associated with.

            cert_manager (reviewboard.certs.manager.CertificateManager,
                          optional):
                A specific Certificate Manager instance.

                If not provided, the default will be used.

            check_hostname (bool, optional):
                Whether to verify that the hostname in the URL matches
                the hostname in the certificate.

            extra_cert_data (str, optional):
                Optional PEM-formatted certificate data to use for
                verification.
        """
        self._cert_manager = cert_manager or default_cert_manager
        self._local_site = local_site
        self._extra_cert_data = extra_cert_data
        self._rb_check_hostname = check_hostname

        # On Python 3.12+, the parent builds a default SSL context if one
        # isn't provided, loading the system certificates in the process.
        # We build our own context per-request in https_open(), so pass a
        # placeholder to avoid that wasted work.
        super().__init__(context=ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT))

    def https_open(
        self,
        req: Request,
    ) -> HTTPResponse:
        """Open an HTTPS connection, converting SSL errors if needed.

        Args:
            req (urllib.request.Request):
                The request to open.

        Returns:
            http.client.HTTPResponse:
            The response from the server.

        Raises:
            reviewboard.certs.errors.CertificateVerificationError:
                An SSL certificate verification error occurred.

            Exception:
                A non-SSL URL error occurred.
        """
        parsed_url = urlparse(req.full_url)
        hostname = parsed_url.hostname or ''
        port = parsed_url.port or 443

        if hostname:
            context = self._cert_manager.build_ssl_context(
                hostname=hostname,
                port=port,
                local_site=self._local_site,
            )

            context.check_hostname = self._rb_check_hostname

            if extra_cert_data := self._extra_cert_data:
                try:
                    context.load_verify_locations(cadata=extra_cert_data)
                except ssl.SSLError as e:
                    logger.error(
                        'Failed to load legacy certificate data into HTTPS '
                        'SSL context for %s:%s: %s',
                        hostname, port, e,
                    )
        else:
            context = None

        self._context = context

        # Specially handle SSLErrors that come from the request, and let
        # anything else bubble up.
        try:
            return super().https_open(req)
        except URLError as e:
            reason = e.reason

            if isinstance(reason, ssl.SSLError):
                self._process_ssl_error(error=reason,
                                        hostname=hostname,
                                        port=port)

            raise

    def _process_ssl_error(
        self,
        *,
        error: ssl.SSLError,
        hostname: str,
        port: int,
    ) -> None:
        """Process the SSL error and raise a suitable verification exception.

        Args:
            error (ssl.SSLError):
                The error to process.

            hostname (str):
                The hostname that was attempted.

            port (int):
                The port that was attempted.

        Raises:
            reviewboard.certs.errors.CertificateVerificationError:
                An SSL certificate verification error occurred.
        """
        # This is an SSL error. Convert it to a
        # CertificateVerificationError.
        if isinstance(error, ssl.SSLCertVerificationError):
            code = (
                CertificateVerificationFailureCode
                .for_ssl_verify_code(error.verify_code)
            )
            detail_msg = error.verify_message
        else:
            code = CertificateVerificationFailureCode.OTHER
            detail_msg = str(error)

        # Fetch the certificate from the server.
        #
        # There's a non-zero chance that it won't be the correct
        # signature (it may have been a temporary glitch or some
        # misconfigured server in a fleet), but we can't get the cert
        # from the exception.
        certificate = Certificate.create_from_server(
            hostname=hostname,
            port=port,
        )

        raise CertificateVerificationError(
            code=code,
            certificate=certificate,
            detail_msg=detail_msg,
        )


def urlopen(
    url: str | Request,
    data: (_DataType | None) = None,
    timeout: Unsettable[float | None] = UNSET,
    *,
    local_site: (LocalSite | None) = None,
    cert_manager: (CertificateManager | None) = None,
) -> _UrlopenRet:
    """A wrapper around urlopen that supports certificate management.

    This is a simple, straight-forward compatibility wrapper around
    :py:func:`urllib.request.urlopen` that makes use of a custom HTTPS
    handler for certificate management and error reporting.

    An HTTPS handler backed by the certificate manager is always added to the
    request so that HTTP to HTTPS redirects behave as expected.

    Callers that need more advanced functionality (such as those that build
    their own list of handlers) should not use this, and instead should
    make use of :py:class:`CertificateVerificationHTTPSHandler`.

    Version Added:
        8.1

    Args:
        url (str or urllib.request.Request):
            The URL to connect to or the request to issue.

        data (object, optional):
            Data to pass to the request.

        timeout (float, optional):
            An explicit timeout for the request.

        local_site (reviewboard.site.models.LocalSite, optional):
            The Local Site the certificates would be associated with.

        cert_manager (reviewboard.certs.manager.CertificateManager,
                      optional):
            A specific Certificate Manager instance.

            If not provided, the default will be used.

    Returns:
        urllib.request._UrlopenRet:
        The response data for the request.
    """
    # Build the keyword arguments to pass to the opener or to urlopen().
    urlopen_kwargs: KwargsDict = {
        'data': data,
    }

    if timeout is not UNSET:
        urlopen_kwargs['timeout'] = timeout

    # Build the HTTPS handler used for certificate lookup.
    opener = build_opener(CertificateVerificationHTTPSHandler(
        cert_manager=cert_manager or default_cert_manager,
        local_site=local_site,
    ))

    # Perform the request.
    return opener.open(url, **urlopen_kwargs)



from __future__ import annotations

from typing import Generic, TYPE_CHECKING

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from haystack import indexes
from typing_extensions import TypeVar

from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Final


_TModel = TypeVar('_TModel',
                  bound=Model,
                  default=Model)


class BaseSearchIndex(Generic[_TModel], indexes.SearchIndex):
    """Base class for a search index.

    This sets up a few common fields we want all indexes to include.

    Version Changed:
        8.0:
        This now supports a generic type for the indexed model.
    """

    #: The model to index.
    model: type[_TModel]

    #: The local site attribute on the model.
    #:
    #: For ForeignKeys, this should be the name of the ID field, as in
    #: 'local_site_id'. For ManyToManyFields, it should be the standard field
    #: name.
    local_site_attr: (str | None) = None

    # Common fields for all search indexes.
    text = indexes.CharField(document=True, use_template=True)
    local_sites = indexes.MultiValueField(null=True)

    #: A constant indicating no Local Site ID on an object.
    NO_LOCAL_SITE_ID: Final[int] = 0

    def get_model(self) -> type[_TModel]:
        """Return the model for this index.

        Returns:
            type:
            The type of model.

        Raises:
            AttributeError:
                :py:attr:`model` was not set on the subclass.
        """
        model = getattr(self, 'model', None)

        if model is None:
            raise AttributeError(
                f'{self.__class__.__name__}.model must be set.'
            )

        return model

    def prepare_local_sites(
        self,
        obj: _TModel,
    ) -> Sequence[str]:
        """Prepare the list of local sites for the search index.

        This will take any associated local sites on the object and store
        them in the index as a list. The search view can then easily look up
        values in the list, regardless of the type of object.

        If the object is not a part of a local site, the list will be
        ``[0]``, indicating no local site.

        Args:
            obj (django.db.models.Model):
                The model instance to prepare.

        Returns:
            list of str:
            The list of of model Local Site IDs as strings, or ``[0]`` if
            there are no Local Sites.

        Raises:
            django.core.exceptions.ImproperlyConfigured:
                One or more attributes were not set on the subclass.
        """
        if not self.local_site_attr:
            raise ImproperlyConfigured('local_site_attr must be set on %r'
                                       % self.__class__)

        if not hasattr(obj, self.local_site_attr):
            raise ImproperlyConfigured(
                '"%s" is not a valid local site attribute on %r'
                % (self.local_site_attr, obj.__class__))

        if not LocalSite.objects.has_local_sites():
            # There aren't any Local Sites on the server, so there's nothing
            # to process.
            return [str(self.NO_LOCAL_SITE_ID)]

        local_sites = getattr(obj, self.local_site_attr, None)

        if self.local_site_attr.endswith('_id'):
            # This is from a ForeignKey. We're working with a numeric ID.
            if local_sites is not None:
                results = [local_sites]
            else:
                results = [self.NO_LOCAL_SITE_ID]
        else:
            # This is most likely a ManyToManyField. Anything else is an
            # error.
            #
            # We want to loop through the actual entries and not the primary
            # keys. The caller is responsible for doing a prefetch_related().
            results = [
                local_site.pk
                for local_site in local_sites.all()
            ] or [self.NO_LOCAL_SITE_ID]

        # Convert these all to strings. This is what MultiValueField would
        # normally do if we didn't prepare it, and is needed for the kinds of
        # comparisons we perform when using Elasticsearch 7.x+.
        return [str(_pk) for _pk in results]



"""Base helpers for WebAPI unit tests."""

from __future__ import annotations

import os
from typing import Optional, TYPE_CHECKING

from django.conf import settings
from django.contrib.auth.models import User
from djblets.siteconfig.models import SiteConfiguration
from djblets.webapi.testing.testcases import WebAPITestCaseMixin

from reviewboard.accounts.backends.standard import StandardAuthBackend
from reviewboard.notifications.tests.mixins import EmailTestHelper
from reviewboard.testing import TestCase
from reviewboard.webapi.tests.mimetypes import error_mimetype

if TYPE_CHECKING:
    from typelets.funcs import KwargsDict
    from typelets.json import JSONDict


_auth_backend_path = '%s.%s' % (StandardAuthBackend.__module__,
                                StandardAuthBackend.__name__)


class BaseWebAPITestCase(WebAPITestCaseMixin, EmailTestHelper, TestCase):
    """Base class for WebAPI unit tests.

    This manages initial setup and teardown for unit tests, and provides
    useful utility functions to aid in testing the API.
    """

    #: The base URL to use for absolute URLs.
    #:
    #: Type:
    #:     str:
    base_url = 'http://testserver'

    #: The default error mimetype.
    #:
    #: Type:
    #:     str
    error_mimetype = error_mimetype

    ######################
    # Instance variables #
    ######################

    #: The site configuration used during the API test.
    #:
    #: Type:
    #:     djblets.siteconfig.models.SiteConfiguration
    siteconfig: SiteConfiguration

    #: The currently logged-in user for the test.
    #:
    #: This may be ``None``.
    #:
    #: Type:
    #:     django.contrib.auth.models.User
    user: Optional[User]

    #: Site configuration settings saved prior to the test.
    #:
    #: Type:
    #:     dict
    _saved_siteconfig_settings: JSONDict

    def setUp(self) -> None:
        """Set up the unit test.

        This will temporarily override site configuration settings to block
        all review e-mail code and disable site-wide login requirements. It
        will then set up a user for the test.

        Settings will be restored once the test concludes.
        """
        super().setUp()

        siteconfig = SiteConfiguration.objects.get_current()
        self._saved_siteconfig_settings = siteconfig.settings.copy()

        siteconfig.set('mail_send_review_mail', False)
        siteconfig.set('auth_require_sitewide_login', False)
        siteconfig.save()

        self.siteconfig = siteconfig

        if 'test_users' in (getattr(self, 'fixtures', None) or []):
            self.user = self._login_user()
        else:
            self.user = None

    def tearDown(self) -> None:
        """Tear down the unit test.

        This will restore any previously-changed settings and clear out
        test state.
        """
        super().tearDown()

        siteconfig = self.siteconfig
        siteconfig.settings = self._saved_siteconfig_settings
        siteconfig.save()

        # Clear out state so we don't leak anything.
        #
        # Skip type checking on the siteconfig since we guarantee it's always
        # set when actually used by a test.
        self._saved_siteconfig_settings = None  # type: ignore
        self.siteconfig = None  # type: ignore
        self.user = None

    def _testHttpCaching(
        self,
        url: str,
        *,
        check_etags: bool = False,
        check_last_modified: bool = False,
    ) -> None:
        """Run a test to check for HTTP caching headers.

        Args:
            url (str):
                The URL to request.

            check_etags (bool, optional):
                Whether to check :mailheader:`ETag` headers.

            check_last_modified (bool, optional):
                Whether to check :mailheader:`Last-Modified` headers.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        client = self.client
        response = client.get(url)

        self.assertHttpOK(response, check_etag=check_etags,
                          check_last_modified=check_last_modified)

        headers: KwargsDict = {}

        if check_etags:
            headers['HTTP_IF_NONE_MATCH'] = response['ETag']

        if check_last_modified:
            headers['HTTP_IF_MODIFIED_SINCE'] = response['Last-Modified']

        response = client.get(url, **headers)

        self.assertHttpNotModified(response)

    #
    # Some utility functions shared across test suites.
    #
    def _login_user(
        self,
        *,
        admin: bool = False,
        local_site: bool = False,
    ) -> User:
        """Log in a user and return it for the test.

        The proper user will be created based on whether a valid LocalSite
        user is needed, and/or an admin user is needed.

        Args:
            admin (bool, optional):
                Whether to log in as an administrator.

                If ``local_site=True`` is also passed, then this will log in
                as a standard user set to be the Local Site's administrator.

            local_site (bool, optional):
                Whether to log in to the default test Local Site.

        Returns:
            django.contrib.auth.models.User:
            The logged-in user.
        """
        user: Optional[User] = None
        username: str

        if local_site:
            # We'll use "doc" unconditionally, and just add to the list of
            # admins if we need doc to be an admin user.
            #
            # In the future, we may want to create a new fixture user that
            # is guaranteed an admin, but this will require updating a number
            # of tests and may be performance implications for the full test
            # suite.
            username = 'doc'

            if admin:
                user = User.objects.get(username=username)
                cur_local_site = self.get_local_site(name=self.local_site_name)
                cur_local_site.admins.add(user)
        else:
            if admin:
                username = 'admin'
            else:
                # Pick a user that's not part of the default LocalSite.
                username = 'grumpy'

        if user is None:
            user = User.objects.get(username=username)

        self.client.force_login(user=user,
                                backend=_auth_backend_path)

        return user

    def get_sample_image_filename(self) -> str:
        """Return the path to a local image file that can be used for tests.

        This must only be used for reading, and not for writing.

        Returns:
            str:
            The absolute file path.
        """
        return os.path.join(settings.STATIC_ROOT, 'rb', 'images', 'logo.png')
